#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import List, Any, Dict
import logging
import time
import os
from celery import Celery, Task
from celery.worker.request import Request
from celery.signals import after_setup_task_logger, after_setup_logger, worker_process_init, \
    celeryd_after_setup, worker_process_shutdown
from celery.utils import worker_direct
from celery.exceptions import SoftTimeLimitExceeded
from billiard.exceptions import WorkerLostError
from mongoengine import Document, StringField, ListField, DateTimeField, ValidationError
from mongoengine.connection import ConnectionFailure
from mongoengine.base.metaclasses import TopLevelDocumentMetaclass
from datetime import datetime
import functools

from nomad import config, utils, infrastructure
import nomad.patch  # pylint: disable=unused-import


if config.logstash.enabled:
    from nomad.utils import structlogging

    def initialize_logstash(logger=None, loglevel=logging.DEBUG, **kwargs):
        structlogging.add_logstash_handler(logger)
        return logger

    after_setup_task_logger.connect(initialize_logstash)
    after_setup_logger.connect(initialize_logstash)


@worker_process_init.connect
def setup(**kwargs):
    # each subprocess is supposed disconnect connect again: https://jira.mongodb.org/browse/PYTHON-2090
    try:
        from mongoengine import disconnect
        disconnect()
    except Exception:
        pass

    infrastructure.setup()
    utils.get_logger(__name__).info(
        'celery configured with acks_late=%s' % str(config.celery.acks_late))


worker_hostname = None


@celeryd_after_setup.connect
def capture_worker_name(sender, instance, **kwargs):
    global worker_hostname
    worker_hostname = sender


@worker_process_shutdown.connect
def on_worker_process_shutdown(*args, **kwargs):
    # We need to make sure not to leave open sessions: https://jira.mongodb.org/browse/PYTHON-2090
    from mongoengine.connection import disconnect
    disconnect()


app = Celery('nomad.processing', broker=config.rabbitmq_url())
app.conf.update(worker_hijack_root_logger=False)
app.conf.update(worker_max_memory_per_child=config.celery.max_memory)
if config.celery.routing == config.CELERY_WORKER_ROUTING:
    app.conf.update(worker_direct=True)

app.conf.task_queue_max_priority = 10

CREATED = 'CREATED'
PENDING = 'PENDING'
RUNNING = 'RUNNING'
FAILURE = 'FAILURE'
SUCCESS = 'SUCCESS'

PROCESS_CALLED = 'CALLED'
PROCESS_RUNNING = 'RUNNING'
PROCESS_COMPLETED = 'COMPLETED'


class InvalidId(Exception): pass


class ProcNotRegistered(Exception): pass


class ProcessAlreadyRunning(Exception): pass


class ProcObjectDoesNotExist(Exception): pass


class ProcMetaclass(TopLevelDocumentMetaclass):
    def __new__(cls, name, bases, attrs):
        cls = super().__new__(cls, name, bases, attrs)

        tasks = []
        setattr(cls, 'tasks', tasks)

        for name, attr in attrs.items():
            task = getattr(attr, '__task_name', None)
            if task is not None and task not in tasks:
                tasks.append(task)

        return cls


class Proc(Document, metaclass=ProcMetaclass):
    '''
    Base class for objects that are involved in processing and need persistent processing
    state.

    It solves two issues. First, distributed operation (via celery) and second keeping
    state of a chain of potentially failing processing tasks. Both are controlled via
    decorators @process and @task. Subclasses should use these decorators on their
    methods. Parameters are not supported for decorated functions. Use fields on the
    document instead.

    Processing state will be persistet at appropriate
    times and must not be persistet manually. All attributes are stored to mongodb.

    Possible processing states are PENDING, RUNNING, FAILURE, and SUCCESS.

    Attributes:
        current_task: the currently running or last completed task
        tasks_status: the overall status of the processing
        errors: a list of errors that happened during processing. Error fail a processing
            run
        warnings: a list of warnings that happened during processing. Warnings do not
            fail a processing run
        create_time: the time of creation (not the start of processing)
        complete_time: the time that processing completed (successfully or not)
        current_process: the currently or last run asyncronous process
        process_status: the status of the currently or last run asyncronous process
    '''

    meta: Any = {
        'abstract': True,
    }

    tasks: List[str] = None
    ''' the ordered list of tasks that comprise a processing run '''

    current_task = StringField(default=None)
    tasks_status = StringField(default=CREATED)
    create_time = DateTimeField(required=True)
    complete_time = DateTimeField()

    errors = ListField(StringField())
    warnings = ListField(StringField())
    last_status_message = StringField(default=None)

    current_process = StringField(default=None)
    process_status = StringField(default=None)

    worker_hostname = StringField(default=None)
    celery_task_id = StringField(default=None)

    @property
    def tasks_running(self) -> bool:
        ''' Returns True of the process has failed or succeeded. '''
        return self.tasks_status not in [SUCCESS, FAILURE]

    @property
    def process_running(self) -> bool:
        ''' Returns True of an asynchrounous process is currently running. '''
        return self.process_status is not None and self.process_status != PROCESS_COMPLETED

    @classmethod
    def process_running_mongoengine_query(cls):
        ''' Returns a mongoengine query dict (to be used in objects) to find running processes. '''
        return dict(process_status__in=[PROCESS_CALLED, PROCESS_RUNNING])

    def get_logger(self):
        return utils.get_logger(
            'nomad.processing', task=self.current_task, proc=self.__class__.__name__,
            process=self.current_process, process_status=self.process_status,
            tasks_status=self.tasks_status)

    @classmethod
    def create(cls, **kwargs):
        ''' Factory method that must be used instead of regular constructor. '''
        assert 'tasks_status' not in kwargs, \
            ''' do not set the status manually, its managed '''

        kwargs.setdefault('create_time', datetime.utcnow())
        self = cls(**kwargs)
        if len(cls.tasks) == 0:
            self.tasks_status = SUCCESS
        else:
            self.tasks_status = PENDING if self.current_task is None else RUNNING
        self.save()

        return self

    def reset(
            self, worker_hostname: str = None, force: bool = False,
            tasks_status: str = PENDING):

        ''' Resets the task chain. Assumes there no current running process. '''
        assert not self.process_running or force

        self.current_task = None
        self.process_status = None
        self.tasks_status = tasks_status
        self.errors = []
        self.warnings = []
        self.worker_hostname = worker_hostname

    @classmethod
    def reset_pymongo_update(cls, worker_hostname: str = None):
        ''' Returns a pymongo update dict part to reset calculations. '''
        return dict(
            current_task=None, process_status=None, tasks_status=PENDING, errors=[], warnings=[],
            worker_hostname=worker_hostname)

    @classmethod
    def get_by_id(cls, id: str, id_field: str):
        try:
            obj = cls.objects(**{id_field: id}).first()
        except ValidationError as e:
            raise InvalidId('%s is not a valid id' % id)
        except ConnectionFailure as e:
            raise e

        if obj is None:
            raise KeyError('%s with id %s does not exist' % (cls.__name__, id))

        return obj

    @classmethod
    def get(cls, obj_id):
        return cls.get_by_id(str(obj_id), 'id')

    @staticmethod
    def log(logger, log_level, msg, **kwargs):
        # TODO there seems to be a bug in structlog, cannot use logger.log
        if log_level == logging.ERROR:
            logger.error(msg, **kwargs)
        elif log_level == logging.WARNING:
            logger.warning(msg, **kwargs)
        elif log_level == logging.INFO:
            logger.info(msg, **kwargs)
        elif log_level == logging.DEBUG:
            logger.debug(msg, **kwargs)
        else:
            logger.critical(msg, **kwargs)

    def fail(self, *errors, log_level=logging.ERROR, **kwargs):
        ''' Allows to fail the process. Takes strings or exceptions as args. '''
        assert self.process_running or self.tasks_running, 'Cannot fail a completed process.'

        failed_with_exception = False

        self.tasks_status = FAILURE

        logger = self.get_logger(**kwargs)
        self.errors = []
        for error in errors:
            if isinstance(error, Exception):
                failed_with_exception = True
                self.errors.append('%s: %s' % (error.__class__.__name__, str(error)))
                Proc.log(
                    logger, log_level, 'task failed with exception',
                    exc_info=error, error=str(error))
            else:
                self.errors.append(str(error))

        self.complete_time = datetime.utcnow()

        if not failed_with_exception:
            errors_str = "; ".join([str(error) for error in errors])
            Proc.log(logger, log_level, 'task failed', errors=errors_str)

        self.on_fail()

        logger.info('process failed')
        if len(self.errors) > 0:
            self.last_satus_message = 'ERROR: %s' % self.errors[-1]

        self.save()

    def on_fail(self):
        pass

    def warning(self, *warnings, log_level=logging.WARNING, **kwargs):
        ''' Allows to save warnings. Takes strings or exceptions as args. '''
        assert self.process_running or self.tasks_running

        logger = self.get_logger(**kwargs)

        for warning in warnings:
            warning = str(warning)
            self.warnings.append(warning)
            Proc.log(logger, log_level, 'task with warning', warning=warning)

    def _continue_with(self, task):
        tasks = self.__class__.tasks
        assert task in tasks, 'task %s must be one of the classes tasks %s' % (task, str(tasks))  # pylint: disable=E1135
        if self.current_task is None:
            assert task == tasks[0], "process has to start with first task %s" % tasks[0]  # pylint: disable=E1136
        elif tasks.index(task) <= tasks.index(self.current_task):
            # task is repeated, probably the celery task of the process was reschedule
            # due to prior worker failure
            self.current_task = task
            self.get_logger().error('task is re-run')
            self.save()
            return True
        else:
            assert tasks.index(task) == tasks.index(self.current_task) + 1, \
                "tasks must be processed in the right order"

        if self.tasks_status == FAILURE:
            return False

        if self.tasks_status == PENDING:
            assert self.current_task is None
            assert task == tasks[0]  # pylint: disable=E1136
            self.tasks_status = RUNNING
            self.current_task = task
            self.get_logger().info('started process')
        else:
            self.current_task = task
            self.get_logger().info('task completed successfully')

        self.save()
        return True

    def _complete(self):
        if self.tasks_status != FAILURE:
            assert self.tasks_status == RUNNING, 'Can only complete a running process, process is %s' % self.tasks_status
            self.tasks_status = SUCCESS
            self.complete_time = datetime.utcnow()
            self.on_tasks_complete()
            self.save()
            self.get_logger().info('completed process')

    def on_tasks_complete(self):
        ''' Callback that is called when the list of task are completed '''
        pass

    def on_process_complete(self, process_name):
        ''' Callback that is called when the corrent process completed '''
        pass

    def block_until_complete(self, interval=0.01):
        '''
        Reloads the process constantly until it sees a completed process with finished tasks.
        Should be used with care as it can block indefinitely. Just intended for testing
        purposes.
        '''
        while self.tasks_running or self.process_running:
            time.sleep(interval)
            self.reload()

    def block_until_process_complete(self, interval=0.01):
        '''
        Reloads the process constantly until it sees a completed process. Should be
        used with care as it can block indefinitely. Just intended for testing purposes.
        '''
        while self.process_running:
            time.sleep(interval)
            self.reload()

    @classmethod
    def process_all(cls, func, query: Dict[str, Any], exclude: List[str] = []):
        '''
        Allows to run process functions for all objects on the given query. Calling
        process functions though the func:`process` wrapper might be slow, because
        it causes a save on each call. This function will use a query based update to
        do the same for all objects at once.
        '''

        running_query = dict(cls.process_running_mongoengine_query())
        running_query.update(query)
        if cls.objects(**running_query).first() is not None:
            raise ProcessAlreadyRunning('Tried to call a processing function on an already processing process.')

        cls._get_collection().update_many(query, {'$set': dict(
            current_process=func.__name__,
            process_status=PROCESS_CALLED)})

        for obj in cls.objects(**query).exclude(*exclude):
            obj._run_process(func)

    def _run_process(self, func):
        if hasattr(func, '__process_unwrapped'):
            func = getattr(func, '__process_unwrapped')

        self_id = self.id.__str__()
        cls_name = self.__class__.__name__

        queue = None
        if config.celery.routing == config.CELERY_WORKER_ROUTING and self.worker_hostname is not None:
            queue = worker_direct(self.worker_hostname).name

        priority = config.celery.priorities.get('%s.%s' % (cls_name, func.__name__), 1)

        logger = utils.get_logger(__name__, cls=cls_name, id=self_id, func=func.__name__)
        logger.debug('calling process function', queue=queue, priority=priority)

        return proc_task.apply_async(
            args=[cls_name, self_id, func.__name__],
            queue=queue, priority=priority)

    def __str__(self):
        return 'proc celery_task_id=%s worker_hostname=%s' % (self.celery_task_id, self.worker_hostname)


def task(func):
    '''
    The decorator for tasks that will be wrapped in exception handling that will fail the process.
    The task methods of a :class:`Proc` class/document comprise a sequence
    (order of methods in class namespace) of tasks. Tasks must be executed in that order.
    Completion of the last task, will put the :class:`Proc` instance into the
    SUCCESS state. Calling the first task will put it into RUNNING state. Tasks will
    only be executed, if the process has not yet reached FAILURE state.
    '''
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        try:
            if self.tasks_status == FAILURE:
                return

            self._continue_with(func.__name__)
            try:
                func(self, *args, **kwargs)

            except Exception as e:
                self.fail(e)

            except SystemExit:
                self.fail('unexpected system exit')

            if self.__class__.tasks[-1] == self.current_task and self.tasks_running:
                self._complete()

        except Exception as e:
            # this is very critical and an indicator that the task fail error handling
            # it self failed
            self.get_logger().critical('task wrapper failed with exception', exc_info=e)

    setattr(wrapper, '__task_name', func.__name__)
    return wrapper


def all_subclasses(cls):
    ''' Helper method to calculate set of all subclasses of a given class. '''
    return set(cls.__subclasses__()).union(
        [s for c in cls.__subclasses__() for s in all_subclasses(c)])


all_proc_cls = {cls.__name__: cls for cls in all_subclasses(Proc)}
''' Name dictionary for all Proc classes. '''


class NomadCeleryRequest(Request):
    '''
    A custom celery request class that allows to catch error in the worker main
    thread, which cannot be caught on the worker threads themselves.
    '''

    def _fail(self, event, **kwargs):
        args = self._payload[0]
        # this might be run in the worker main thread, which does not have a mongo
        # connection by default
        if infrastructure.mongo_client is None:
            infrastructure.setup_mongo()

        proc = unwarp_task(self.task, *args)
        proc.fail(event, **kwargs)
        proc.process_status = PROCESS_COMPLETED
        proc.on_process_complete(None)
        proc.save()

    def on_timeout(self, soft, timeout):
        if not soft:
            self._fail('task timeout occurred', timeout=timeout)

        super().on_timeout(soft, timeout)

    def on_failure(self, exc_info, send_failed_event=True, return_ok=False):
        if isinstance(exc_info.exception, WorkerLostError):
            infrastructure.setup()
            utils.get_logger(__name__).error(
                'detected WorkerLostError', exc_info=exc_info.exception)
            self._fail(
                'task failed due to worker lost: %s' % str(exc_info.exception),
                exc_info=exc_info)

        super().on_failure(
            exc_info,
            send_failed_event=send_failed_event,
            return_ok=return_ok
        )


class NomadCeleryTask(Task):
    Request = NomadCeleryRequest


def unwarp_task(task, cls_name, self_id, *args, **kwargs):
    '''
    Retrieves the proc object that the given task is executed on from the database.
    '''
    logger = utils.get_logger(__name__, cls=cls_name, id=self_id)

    # get the process class
    global all_proc_cls
    cls = all_proc_cls.get(cls_name, None)
    if cls is None:
        # refind all Proc classes, since more modules might have been imported by now
        all_proc_cls = {cls.__name__: cls for cls in all_subclasses(Proc)}
        cls = all_proc_cls.get(cls_name, None)

    if cls is None:
        logger.critical('document not a subcass of Proc')
        raise ProcNotRegistered('document %s not a subclass of Proc' % cls_name)

    # get the process instance
    try:
        try:
            self = cls.get(self_id)
        except KeyError as e:
            from nomad.app import flask
            if flask.app.config['TESTING']:
                # This only happens in tests, where it is not always avoidable that
                # tasks from old test-cases bleed over.
                raise ProcObjectDoesNotExist()
            logger.warning('called object is missing, retry')
            raise task.retry(exc=e, countdown=3)
    except KeyError:
        logger.critical('called object is missing, retries exeeded', proc_id=self_id)
        raise ProcObjectDoesNotExist()

    return self


@app.task(
    bind=True, base=NomadCeleryTask, ignore_results=True, max_retries=3,
    acks_late=config.celery.acks_late, soft_time_limit=config.celery.timeout,
    time_limit=config.celery.timeout * 2)
def proc_task(task, cls_name, self_id, func_attr):
    '''
    The celery task that is used to execute async process functions.
    It ignores results, since all results are handled via the self document.
    It retries for 3 times with a countdown of 3 on missing 'selfs', since this
    might happen in sharded, distributed mongo setups where the object might not
    have yet been propagated and therefore appear missing.
    '''
    self = unwarp_task(task, cls_name, self_id)

    logger = self.get_logger()
    logger.debug('received process function call')

    self.worker_hostname = worker_hostname
    self.celery_task_id = task.request.id

    # get the process function
    func = getattr(self, func_attr, None)
    if func is None:
        logger.error('called function not a function of proc class')
        self.fail('called function %s is not a function of proc class %s' % (func_attr, cls_name))
        self.process_status = PROCESS_COMPLETED
        self.save()
        return

    # unwrap the process decorator
    func = getattr(func, '__process_unwrapped', None)
    if func is None:
        logger.error('called function was not decorated with @process')
        self.fail('called function %s was not decorated with @process' % func_attr)
        self.process_status = PROCESS_COMPLETED
        self.on_process_complete(None)
        self.save()
        return

    # call the process function
    deleted = False
    try:
        self.process_status = PROCESS_RUNNING
        os.chdir(config.fs.working_directory)
        with utils.timer(logger, 'process executed on worker'):
            deleted = func(self)
    except SoftTimeLimitExceeded as e:
        logger.error('exceeded the celery task soft time limit')
        self.fail(e)
    except Exception as e:
        self.fail(e)
    except SystemExit as e:
        self.fail(e)
    finally:
        if deleted is None or not deleted:
            self.process_status = PROCESS_COMPLETED
            self.on_process_complete(func.__name__)
            self.save()


def process(func):
    '''
    The decorator for process functions that will be called async via celery.
    All calls to the decorated method will result in celery task requests.
    To transfer state, the instance will be saved to the database and loading on
    the celery task worker. Process methods can call other (process) functions/methods on
    other :class:`Proc` instances. Each :class:`Proc` instance can only run one
    any process at a time.
    '''
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        assert len(args) == 0 and len(kwargs) == 0, 'process functions must not have arguments'
        if self.process_running:
            raise ProcessAlreadyRunning('Tried to call a processing function on an already processing process.')

        self.current_process = func.__name__
        self.process_status = PROCESS_CALLED
        self.save()

        self._run_process(func)

    task = getattr(func, '__task_name', None)
    if task is not None:
        setattr(wrapper, '__task_name', task)

    setattr(wrapper, '__process_unwrapped', func)

    return wrapper



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from nomad.metainfo import Section, Quantity, Package

from nomad.datamodel.metainfo.public import section_single_configuration_calculation as SCC


m_package = Package(
    name='openmx_nomadmetainfo_json',
    description='None')


# We extend the existing common definition of a section "single configuration calculation"
class OpenmxSCC(SCC):
    # We alter the default base class behavior to add all definitions to the existing
    # base class instead of inheriting from the base class
    m_def = Section(extends_base_section=True)

    # We define an additional example quantity. Use the prefix x_<parsername>_ to denote
    # non common quantities.
    x_example_magic_value = Quantity(type=int, description='The magic value from a magic source.')


m_package.__init_metainfo__()



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
import logging
import numpy as np
from ase.data import chemical_symbols
import yaml
try:
    from yaml import CLoader
except Exception:
    pass

from nomad.units import ureg
from nomad.parsing import FairdiParser
from nomad.datamodel.metainfo.common_dft import Run, Method, System, XCFunctionals,\
    SingleConfigurationCalculation, ScfIteration


class BigDFTParser(FairdiParser):
    def __init__(self):
        super().__init__(
            name='parsers/bigdft', code_name='BigDFT', code_homepage='http://bigdft.org/',
            mainfile_contents_re=(
                # r'__________________________________ A fast and precise DFT wavelet code\s*'
                # r'\|     \|     \|     \|     \|     \|\s*'
                # r'\|     \|     \|     \|     \|     \|      BBBB         i       gggggg\s*'
                # r'\|_____\|_____\|_____\|_____\|_____\|     B    B               g\s*'
                # r'\|     \|  :  \|  :  \|     \|     \|    B     B        i     g\s*'
                # r'\|     \|-0\+--\|-0\+--\|     \|     \|    B    B         i     g        g\s*'
                r'\|_____\|__:__\|__:__\|_____\|_____\|___ BBBBB          i     g         g\s*'
                # r'\|  :  \|     \|     \|  :  \|     \|    B    B         i     g         g\s*'
                # r'\|--\+0-\|     \|     \|-0\+--\|     \|    B     B     iiii     g         g\s*'
                # r'\|__:__\|_____\|_____\|__:__\|_____\|    B     B        i      g        g\s*'
                # r'\|     \|  :  \|  :  \|     \|     \|    B BBBB        i        g      g\s*'
                # r'\|     \|-0\+--\|-0\+--\|     \|     \|    B        iiiii          gggggg\s*'
                # r'\|_____\|__:__\|__:__\|_____\|_____\|__BBBBB\s*'
                # r'\|     \|     \|     \|  :  \|     \|                           TTTTTTTTT\s*'
                # r'\|     \|     \|     \|--\+0-\|     \|  DDDDDD          FFFFF        T\s*'
                # r'\|_____\|_____\|_____\|__:__\|_____\| D      D        F        TTTT T\s*'
                # r'\|     \|     \|     \|  :  \|     \|D        D      F        T     T\s*'
                # r'\|     \|     \|     \|--\+0-\|     \|D         D     FFFF     T     T\s*'
                # r'\|_____\|_____\|_____\|__:__\|_____\|D___      D     F         T    T\s*'
                # r'\|     \|     \|  :  \|     \|     \|D         D     F          TTTTT\s*'
                # r'\|     \|     \|--\+0-\|     \|     \| D        D     F         T    T\s*'
                # r'\|_____\|_____\|__:__\|_____\|_____\|          D     F        T     T\s*'
                # r'\|     \|     \|     \|     \|     \|         D               T    T\s*'
                # r'\|     \|     \|     \|     \|     \|   DDDDDD       F         TTTT\s*'
                # r'\|_____\|_____\|_____\|_____\|_____\|______                    www\.bigdft\.org'
            ))
        self.yaml_dict = dict()

        # TODO complete mapping
        self.xc_mapping = {
            1: ['LDA_XC_TETER93'],
            11: ['GGA_C_PBE", "GGA_X_PBE'],
            12: ['GGA_X_PBE'],
            15: ['GGA_C_PBE", "GGA_X_RPBE'],
            16: ['GGA_XC_HCTH_93'],
            17: ['GGA_XC_HCTH_120'],
            26: ['GGA_XC_HCTH_147'],
            27: ['GGA_XC_HCTH_407'],
            100: ['HF_X'],
            # libxc
            "001": "LDA_X",
            "002": "LDA_C_WIGNER",
            "003": "LDA_C_RPA",
            "004": "LDA_C_HL",
            "005": "LDA_C_GL",
            "006": "LDA_C_XALPHA",
            "007": "LDA_C_VWN",
            "008": "LDA_C_VWN_RPA",
            "009": "LDA_C_PZ",
            "010": "LDA_C_PZ_MOD",
            "011": "LDA_C_OB_PZ",
            "012": "LDA_C_PW",
            "013": "LDA_C_PW_MOD",
            "014": "LDA_C_OB_PW",
            "015": "LDA_C_2D_AMGB",
            "016": "LDA_C_2D_PRM",
            "017": "LDA_C_vBH",
            "018": "LDA_C_1D_CSC",
            "019": "LDA_X_2D",
            "020": "LDA_XC_TETER93",
            "021": "LDA_X_1D",
            "101": "GGA_X_PBE",
            "102": "GGA_X_PBE_R",
            "103": "GGA_X_B86",
            "104": "GGA_X_B86_R",
            "105": "GGA_X_B86_MGC",
            "106": "GGA_X_B88",
            "107": "GGA_X_G96",
            "108": "GGA_X_PW86",
            "109": "GGA_X_PW91",
            "110": "GGA_X_OPTX",
            "111": "GGA_X_DK87_R1",
            "112": "GGA_X_DK87_R2",
            "113": "GGA_X_LG93",
            "114": "GGA_X_FT97_A",
            "115": "GGA_X_FT97_B",
            "116": "GGA_X_PBE_SOL",
            "117": "GGA_X_RPBE",
            "118": "GGA_X_WC",
            "119": "GGA_X_mPW91",
            "120": "GGA_X_AM05",
            "121": "GGA_X_PBEA",
            "122": "GGA_X_MPBE",
            "123": "GGA_X_XPBE",
            "124": "GGA_X_2D_B86_MGC",
            "125": "GGA_X_BAYESIAN",
            "126": "GGA_X_PBE_JSJR",
            "127": "GGA_X_2D_B88",
            "128": "GGA_X_2D_B86",
            "129": "GGA_X_2D_PBE",
            "130": "GGA_C_PBE",
            "131": "GGA_C_LYP",
            "132": "GGA_C_P86",
            "133": "GGA_C_PBE_SOL",
            "134": "GGA_C_PW91",
            "135": "GGA_C_AM05",
            "136": "GGA_C_XPBE",
            "137": "GGA_C_LM",
            "138": "GGA_C_PBE_JRGX",
            "139": "GGA_X_OPTB88_VDW",
            "140": "GGA_X_PBEK1_VDW",
            "141": "GGA_X_OPTPBE_VDW",
            "160": "GGA_XC_LB",
            "161": "GGA_XC_HCTH_93",
            "162": "GGA_XC_HCTH_120",
            "163": "GGA_XC_HCTH_147",
            "164": "GGA_XC_HCTH_407",
            "165": "GGA_XC_EDF1",
            "166": "GGA_XC_XLYP",
            "167": "GGA_XC_B97",
            "168": "GGA_XC_B97_1",
            "169": "GGA_XC_B97_2",
            "170": "GGA_XC_B97_D",
            "171": "GGA_XC_B97_K",
            "172": "GGA_XC_B97_3",
            "173": "GGA_XC_PBE1W",
            "174": "GGA_XC_MPWLYP1W",
            "175": "GGA_XC_PBELYP1W",
            "176": "GGA_XC_SB98_1a",
            "177": "GGA_XC_SB98_1b",
            "178": "GGA_XC_SB98_1c",
            "179": "GGA_XC_SB98_2a",
            "180": "GGA_XC_SB98_2b",
            "181": "GGA_XC_SB98_2c",
            "401": "HYB_GGA_XC_B3PW91",
            "402": "HYB_GGA_XC_B3LYP",
            "403": "HYB_GGA_XC_B3P86",
            "404": "HYB_GGA_XC_O3LYP",
            "405": "HYB_GGA_XC_mPW1K",
            "406": "HYB_GGA_XC_PBEH",
            "407": "HYB_GGA_XC_B97",
            "408": "HYB_GGA_XC_B97_1",
            "410": "HYB_GGA_XC_B97_2",
            "411": "HYB_GGA_XC_X3LYP",
            "412": "HYB_GGA_XC_B1WC",
            "413": "HYB_GGA_XC_B97_K",
            "414": "HYB_GGA_XC_B97_3",
            "415": "HYB_GGA_XC_mPW3PW",
            "416": "HYB_GGA_XC_B1LYP",
            "417": "HYB_GGA_XC_B1PW91",
            "418": "HYB_GGA_XC_mPW1PW",
            "419": "HYB_GGA_XC_mPW3LYP",
            "420": "HYB_GGA_XC_SB98_1a",
            "421": "HYB_GGA_XC_SB98_1b",
            "422": "HYB_GGA_XC_SB98_1c",
            "423": "HYB_GGA_XC_SB98_2a",
            "424": "HYB_GGA_XC_SB98_2b",
            "425": "HYB_GGA_XC_SB98_2c",
            "201": "MGGA_X_LTA",
            "202": "MGGA_X_TPSS",
            "203": "MGGA_X_M06L",
            "204": "MGGA_X_GVT4",
            "205": "MGGA_X_TAU_HCTH",
            "206": "MGGA_X_BR89",
            "207": "MGGA_X_BJ06",
            "208": "MGGA_X_TB09",
            "209": "MGGA_X_RPP09",
            "231": "MGGA_C_TPSS",
            "232": "MGGA_C_VSXC",
            "301": "LCA_OMC",
            "302": "LCA_LCH",
        }

    def _extract(self, name, source, default=None):
        for key in source.keys():
            if key.lower() == name.lower():
                return source.pop(key)
        return default

    def parse_method(self):
        sec_method = self.archive.section_run[0].m_create(Method)
        sec_method.electronic_structure_method = 'DFT'

        data = self._extract('dft', self.yaml_dict, {})
        sec_method.total_charge = self._extract('qcharge', data, 0)
        sec_method.scf_max_iteration = self._extract('itermax', data, 0)
        sec_method.number_of_spin_channels = self._extract('nspin', data, 1)

        xc_id = self._extract('xc id', data)
        if xc_id is None:
            xc_id = self._extract('ixc', data)
        if xc_id < 0:
            # libxc
            xc_id = '%06d' % abs(xc_id)
            xc_functionals = [self.xc_mapping.get(xc_id[:3]), self.xc_mapping.get(xc_id[3:])]
        else:
            xc_functionals = self.xc_mapping.get(xc_id, [])

        xc_functionals = [f for f in xc_functionals if f is not None]
        for functional in xc_functionals:
            sec_xc_functional = sec_method.m_create(XCFunctionals)
            sec_xc_functional.XC_functional_name = functional
        sec_method.XC_functional = '_'.join(xc_functionals)

    def parse_system(self):
        sec_system = self.archive.section_run[0].m_create(System)

        data = self._extract('Atomic structure', self.yaml_dict, {})
        labels = []
        positions = []
        for atom in self._extract('positions', data, []):
            for label, position in atom.items():
                if label in chemical_symbols:
                    # some entries may not be symbols
                    labels.append(label)
                    positions.append(position)
        if labels:
            sec_system.atom_positions = positions * ureg.angstrom
            sec_system.atom_labels = labels

        cell = self._extract('cell', data)
        if cell is None:
            data = self._extract('Sizes of the simulation domain', self.yaml_dict, {})
            cell = self._extract('angstroem', data)
        if cell is not None:
            sec_system.lattice_vectors = np.diag(cell) * ureg.angstrom
            sec_system.configuration_periodic_dimensions = [True, True, True]

        data = self._extract('Atomic System Properties', self.yaml_dict, {})
        sec_system.number_of_atoms = self._extract('number of atoms', data, 0)

        pbc = self._extract('boundary conditions', data, 'Periodic').lower()
        if pbc == 'free':
            sec_system.configuration_periodic_dimensions = [False, False, False]
        elif pbc == 'periodic':
            sec_system.configuration_periodic_dimensions = [True, True, True]
        elif pbc == 'surface':
            sec_system.configuration_periodic_dimensions = [True, False, True]

    def parse_scc(self):
        sec_scc = self.archive.section_run[0].m_create(SingleConfigurationCalculation)

        energy = self._extract('Energy (Hartree)', self.yaml_dict)
        if energy is not None:
            sec_scc.energy_total = energy * ureg.hartree

        forces = self._extract('Atomic Forces (Ha/Bohr)', self.yaml_dict)
        if forces is not None:
            sec_scc.atom_forces = [list(f.values())[0] for f in forces] * (ureg.hartree / ureg.bohr)

        data = self._extract('Ground State Optimization', self.yaml_dict, [{}])
        hamiltonian = self._extract('hamiltonian optimization', data[0])
        if hamiltonian is None:
            return

        energy_mapping = {
            'exc': 'energy_XC', 'evxc': 'energy_XC_potential',
            'eh': 'energy_correction_hartree', 'ekin': 'electronic_kinetic_energy',
            'eks': 'energy_total', 'd': 'energy_change'}

        subspace = self._extract('subspace optimization', hamiltonian[0], {})
        wavefunction = self._extract('wavefunctions iterations', subspace, [])
        sec_scc.number_of_scf_iterations = len(wavefunction)
        for iteration in wavefunction:
            sec_scf = sec_scc.m_create(ScfIteration)
            energies = self._extract('energies', iteration, {})
            for key, val in energies.items():
                key = energy_mapping.get(key.lower())
                if key is not None:
                    setattr(sec_scf, '%s_scf_iteration' % key, val * ureg.hartree)

            for key, val in iteration.items():
                key = energy_mapping.get(key.lower())
                if key is not None:
                    setattr(sec_scf, '%s_scf_iteration' % key, val * ureg.hartree)

    def parse(self, filepath, archive, logger):
        self.filepath = os.path.abspath(filepath)
        self.archive = archive
        self.logger = logger if logger is not None else logging.getLogger(__name__)

        with open(self.filepath) as f:
            try:
                # we simply load everything at once as the metainfo framework does not
                # support streaming
                self.yaml_dict = yaml.load(f, Loader=CLoader)
            except Exception:
                try:
                    self.yaml_dict = yaml.safe_load(f)
                except Exception:
                    self.logger.error('Error loading yaml file.')
                    return

        sec_run = self.archive.m_create(Run)
        sec_run.program_name = 'BigDFT'
        sec_run.program_basis_set_type = 'real-space grid'
        sec_run.program_version = str(self.yaml_dict.pop('Version Number', ''))

        self.parse_method()
        self.parse_system()
        self.parse_scc()

        sec_scc = sec_run.section_single_configuration_calculation[-1]
        sec_scc.single_configuration_calculation_to_system_ref = sec_run.section_system[-1]
        sec_scc.single_configuration_to_calculation_method_ref = sec_run.section_method[-1]



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import re
import numpy as np

from nomad.parsing.file_parser import BasicParser
from nomad.units import ureg


class LibAtomsParser(BasicParser):
    def __init__(self):
        re_f = r'\-*\d+\.\d+E*e*\-*\+*\d*'

        super().__init__(
            specifications=dict(
                name='parsers/lib-atoms', code_name='libAtoms', code_homepage='https://libatoms.github.io/',
                mainfile_contents_re=(r'\s*<GAP_params\s')),
            units_mapping=dict(length=ureg.angstrom, energy=ureg.eV),
            # necessary to include version due to auto type conversion
            program_version=r'(svn\_version\=\"\d+\")',
            atom_labels_atom_positions_atom_forces=r'(\<\!\[CDATA\[[A-Z][a-z]* +[\s\S]+?)(?:\<\!\[CDATA\[\d+\]\]\>|\Z)',
            energy_total=rf'slice\_sample energy\=({re_f})')



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import Dict
from elasticsearch_dsl import Q, Date
from cachetools import cached

from optimade.filterparser import LarkParser
from optimade.filtertransformers.elasticsearch import (
    Quantity, ElasticTransformer as OPTElasticTransformer)
from optimade.models import CHEMICAL_SYMBOLS, ATOMIC_NUMBERS

from .common import provider_specific_fields


_parser = LarkParser(version=(0, 10, 1))


class FilterException(Exception):
    ''' Raised on parsing a filter expression with syntactic of semantic errors. '''
    pass


@cached(cache={})
def _get_transformer(nomad_properties, without_prefix):
    from nomad.datamodel import OptimadeEntry
    quantities: Dict[str, Quantity] = {
        q.name: Quantity(
            q.name, es_field='dft.optimade.%s' % q.name,
            elastic_mapping_type=q.a_search.mapping.__class__)

        for q in OptimadeEntry.m_def.all_quantities.values()
        if 'search' in q.m_annotations}

    quantities['id'] = Quantity('id', es_field='calc_id')
    quantities['immutable_id'] = Quantity('immutable_id', es_field='calc_id')
    quantities['last_modified'] = Quantity(
        'last_modified', es_field='upload_time', elastic_mapping_type=Date)

    quantities['elements'].length_quantity = quantities['nelements']
    quantities['elements'].has_only_quantity = Quantity(name='only_atoms')
    quantities['elements'].nested_quantity = quantities['elements_ratios']
    quantities['elements_ratios'].nested_quantity = quantities['elements_ratios']

    if nomad_properties is not None:
        for name, search_quantity in provider_specific_fields():
            names = ['_nmd_' + name]
            if without_prefix:
                names.append(name)

            for name in names:
                if name not in quantities:
                    quantities[name] = Quantity(
                        name,
                        es_field=search_quantity.search_field,
                        elastic_mapping_type=search_quantity.mapping.__class__)

    return ElasticTransformer(quantities=quantities.values())


def parse_filter(filter_str: str, nomad_properties='dft', without_prefix=False) -> Q:
    ''' Parses the given optimade filter str and returns a suitable elastic search query.

    Arguments:
        filter_str: Can be direct user input with no prior processing.
        nomad_properties: Also include the nomad proprietary properties of the given domain.
        without_prefix: Do not prefix the nomad proprietary properties with _nmd_.

    Raises:
        FilterException: If the given str cannot be parsed, or if there are any semantic
            errors in the given expression.
    '''
    transformer = _get_transformer(nomad_properties, without_prefix)

    try:
        parse_tree = _parser.parse(filter_str)
    except Exception as e:
        raise FilterException('Syntax error: %s' % str(e))

    try:
        query = transformer.transform(parse_tree)
    except Exception as e:
        raise FilterException('Semantic error: %s' % str(e))

    return query


class ElasticTransformer(OPTElasticTransformer):
    def _has_query_op(self, quantities, op, predicate_zip_list):
        # We override this to add 'HAS ONLY' support.
        if op == 'HAS ONLY':
            # HAS ONLY comes with heavy limitations, because there is no such thing
            # in elastic search. Only supported for elements, where we can construct
            # an anonymous 'formula' based on elements sorted by order number and
            # can do a = comparision to check if all elements are contained
            if len(quantities) > 1:
                raise Exception('HAS ONLY is not supported with zip')
            quantity = quantities[0]

            if quantity.has_only_quantity is None:
                raise Exception('HAS ONLY is not supported by %s' % quantity.name)

            def values():
                for predicates in predicate_zip_list:
                    if len(predicates) != 1:
                        raise Exception('Tuples not supported in HAS ONLY')
                    op, value = predicates[0]
                    if op != '=':
                        raise Exception('Predicated not supported in HAS ONLY')
                    if not isinstance(value, str):
                        raise Exception('Only strings supported in HAS ONLY')
                    yield value

            try:
                order_numbers = list([ATOMIC_NUMBERS[element] for element in values()])
                order_numbers.sort()
                value = ''.join(
                    [CHEMICAL_SYMBOLS[number - 1] for number in order_numbers]
                )
            except KeyError:
                raise NotImplementedError('HAS ONLY is only supported for chemical symbols')

            return Q('term', **{quantity.has_only_quantity.name: value})

        else:
            return super()._has_query_op(quantities, op, predicate_zip_list)

    def property_zip_addon(self, args):
        return args

    def value_zip(self, args):
        return self.value_list(args)

    def value_zip_list(self, args):
        return args



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import typing
import os
import json
import click
import sys

from nomad import utils, parsing, normalizing, datamodel

from .cli import cli


def parse(
        mainfile_path: str,
        parser_name: str = None,
        backend_factory: typing.Callable = None,
        strict: bool = True, logger=None):
    '''
    Run the given parser on the downloaded calculation. If no parser is given,
    do parser matching and use the respective parser.
    '''
    from nomad.parsing import parsers
    mainfile = os.path.basename(mainfile_path)

    if logger is None:
        logger = utils.get_logger(__name__)
    if parser_name is not None:
        parser = parsers.parser_dict.get(parser_name)
        assert parser is not None, 'the given parser must exist'
    else:
        parser = parsers.match_parser(mainfile_path, strict=strict)
        if isinstance(parser, parsing.MatchingParser):
            parser_name = parser.name
        else:
            parser_name = parser.__class__.__name__

    assert parser is not None, 'there is no parser matching %s' % mainfile
    logger = logger.bind(parser=parser.name)  # type: ignore
    logger.info('identified parser')
    if hasattr(parser, 'backend_factory'):
        setattr(parser, 'backend_factory', backend_factory)

    entry_archive = datamodel.EntryArchive()
    metadata = entry_archive.m_create(datamodel.EntryMetadata)
    try:
        cwd = os.getcwd()
        mainfile_path = os.path.abspath(mainfile_path)
        os.chdir(os.path.abspath(os.path.dirname(mainfile_path)))
        parser.parse(mainfile_path, entry_archive, logger=logger)
        os.chdir(cwd)
    except Exception as e:
        logger.error('parsing was not successful', exc_info=e)

    if metadata.domain is None:
        metadata.domain = parser.domain

    logger.info('ran parser')
    return entry_archive


def normalize(
        normalizer: typing.Union[str, typing.Callable], entry_archive, logger=None):

    if logger is None:
        logger = utils.get_logger(__name__)

    if isinstance(normalizer, str):
        normalizer = next(
            normalizer_instance for normalizer_instance in normalizing.normalizers
            if normalizer_instance.__class__.__name__ == normalizer)

    assert normalizer is not None, 'there is no normalizer %s' % str(normalizer)
    normalizer_instance = typing.cast(typing.Callable, normalizer)(entry_archive)
    logger = logger.bind(normalizer=normalizer_instance.__class__.__name__)
    logger.info('identified normalizer')

    normalizer_instance.normalize(logger=logger)
    logger.info('ran normalizer')


def normalize_all(entry_archive, logger=None):
    '''
    Parse the downloaded calculation and run the whole normalizer chain.
    '''
    for normalizer in normalizing.normalizers:
        if normalizer.domain == entry_archive.section_metadata.domain:
            normalize(normalizer, entry_archive, logger=logger)


@cli.command(help='Run parsing and normalizing locally.', name='parse')
@click.argument('MAINFILE', nargs=1, required=True, type=str)
@click.option('--show-archive', is_flag=True, default=False, help='Print the archive data.')
@click.option('--show-metadata', is_flag=True, default=False, help='Print the extracted repo metadata.')
@click.option('--skip-normalizers', is_flag=True, default=False, help='Do not run the normalizer.')
@click.option('--not-strict', is_flag=True, help='Do also match artificial parsers.')
@click.option('--parser', help='Skip matching and use the provided parser')
def _parse(mainfile, show_archive, show_metadata, skip_normalizers, not_strict, parser):
    kwargs = dict(strict=not not_strict, parser_name=parser)

    entry_archive = parse(mainfile, **kwargs)

    if not skip_normalizers:
        normalize_all(entry_archive)
        entry_archive.section_metadata.apply_domain_metadata(entry_archive)

    if show_archive:
        json.dump(entry_archive.m_to_dict(), sys.stdout, indent=2)

    if show_metadata:
        metadata = entry_archive.section_metadata
        metadata.apply_domain_metadata(entry_archive)
        json.dump(metadata.m_to_dict(), sys.stdout, indent=4)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

'''
This module comprises a set of persistent document classes that hold all user related
data. These are information about users, their uploads and datasets, the associated
calculations, and files


.. autoclass:: Calc

.. autoclass:: Upload

'''

from typing import cast, List, Any, Tuple, Iterator, Dict, cast, Iterable
from mongoengine import (
    StringField, DateTimeField, DictField, BooleanField, IntField, ListField)
import logging
from structlog import wrap_logger
from contextlib import contextmanager
import os.path
from datetime import datetime
from pymongo import UpdateOne
import hashlib
from structlog.processors import StackInfoRenderer, format_exc_info, TimeStamper
import yaml
import json
from functools import lru_cache
import urllib.parse
import requests

from nomad import utils, config, infrastructure, search, datamodel, metainfo, parsing
from nomad.files import (
    PathObject, UploadFiles, ExtractError, ArchiveBasedStagingUploadFiles,
    PublicUploadFiles, StagingUploadFiles)
from nomad.processing.base import Proc, process, task, PENDING, SUCCESS, FAILURE
from nomad.parsing.parsers import parser_dict, match_parser
from nomad.normalizing import normalizers
from nomad.datamodel import (
    EntryArchive, EditableUserMetadata, OasisMetadata, UserProvidableMetadata)
from nomad.archive import (
    write_partial_archive_to_mongo, delete_partial_archives_from_mongo)
from nomad.datamodel.encyclopedia import EncyclopediaMetadata


section_metadata = datamodel.EntryArchive.section_metadata.name
section_workflow = datamodel.EntryArchive.section_workflow.name


_editable_metadata: Dict[str, metainfo.Definition] = {}
_editable_metadata.update(**{
    quantity.name: quantity for quantity in UserProvidableMetadata.m_def.definitions})
_editable_metadata.update(**{
    quantity.name: quantity for quantity in EditableUserMetadata.m_def.definitions})

_oasis_metadata = {
    quantity.name: quantity for quantity in OasisMetadata.m_def.definitions}


def _pack_log_event(logger, method_name, event_dict):
    try:
        log_data = dict(event_dict)
        log_data.update(**{
            key: value
            for key, value in getattr(logger, '_context', {}).items()
            if key not in ['service', 'release', 'upload_id', 'calc_id', 'mainfile', 'process_status']})
        log_data.update(logger=logger.name)

        return log_data
    except Exception:
        # raising an exception would cause an indefinite loop
        return event_dict


_log_processors = [
    StackInfoRenderer(),
    _pack_log_event,
    format_exc_info,
    TimeStamper(fmt="%Y-%m-%d %H:%M.%S", utc=False)]


def _normalize_oasis_upload_metadata(upload_id, metadata):
    # This is overwritten by the tests to do necessary id manipulations
    return upload_id, metadata


class Calc(Proc):
    '''
    Instances of this class represent calculations. This class manages the elastic
    search index entry, files, and archive for the respective calculation.

    It also contains the calculations processing and its state.

    The attribute list, does not include the various metadata properties generated
    while parsing, including ``code_name``, ``code_version``, etc.

    Attributes:
        calc_id: the calc_id of this calc
        parser: the name of the parser used to process this calc
        upload_id: the id of the upload used to create this calculation
        mainfile: the mainfile (including path in upload) that was used to create this calc

        metadata: the metadata record wit calc and user metadata, see :class:`datamodel.EntryMetadata`
    '''
    calc_id = StringField(primary_key=True)
    upload_id = StringField()
    mainfile = StringField()
    parser = StringField()

    metadata = DictField()

    meta: Any = {
        'strict': False,
        'indexes': [
            'upload_id',
            'parser',
            ('upload_id', 'mainfile'),
            ('upload_id', 'parser'),
            ('upload_id', 'tasks_status'),
            ('upload_id', 'current_task'),
            ('upload_id', 'process_status'),
            ('upload_id', 'metadata.nomad_version'),
            'metadata.processed',
            'metadata.last_processing',
            'metadata.published',
            'metadata.datasets',
            'metadata.pid'
        ]
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._parser_results: EntryArchive = None
        self._upload: Upload = None
        self._upload_files: ArchiveBasedStagingUploadFiles = None
        self._calc_proc_logs: List[Any] = None

        self._entry_metadata = None

    @classmethod
    def get(cls, id):
        return cls.get_by_id(id, 'calc_id')

    @property
    def entry_id(self) -> str:
        ''' Just an alias for calc_id. '''
        return self.calc_id

    @property
    def mainfile_file(self) -> PathObject:
        return self.upload_files.raw_file_object(self.mainfile)

    @property
    def upload(self) -> 'Upload':
        if not self._upload:
            self._upload = Upload.get(self.upload_id)
            self._upload.worker_hostname = self.worker_hostname
        return self._upload

    def apply_entry_metadata(self, entry_metadata: datamodel.EntryMetadata):
        self.metadata = entry_metadata.m_to_dict(
            include_defaults=True,
            categories=[datamodel.MongoMetadata])  # TODO use embedded doc?

    def create_metadata(self) -> datamodel.EntryMetadata:
        '''
        Returns a :class:`nomad.datamodel.EntryMetadata` with values from this
        processing object, not necessarily the user metadata nor the metadata from
        the archive.
        '''
        entry_metadata = datamodel.EntryMetadata()
        if self.parser is not None:
            parser = parser_dict[self.parser]
            if parser.domain:
                entry_metadata.domain = parser_dict[self.parser].domain
        entry_metadata.upload_id = self.upload_id
        entry_metadata.calc_id = self.calc_id
        entry_metadata.mainfile = self.mainfile
        entry_metadata.nomad_version = config.meta.version
        entry_metadata.nomad_commit = config.meta.commit
        entry_metadata.uploader = self.upload.user_id
        entry_metadata.upload_time = self.upload.upload_time
        entry_metadata.upload_name = self.upload.name

        return entry_metadata

    def entry_metadata(self, upload_files: UploadFiles) -> datamodel.EntryMetadata:
        '''
        Returns a complete set of :class:`nomad.datamodel.EntryMetadata` including
        the user metadata and metadata from the archive.

        Arguments:
            upload_files:
                The :class:`nomad.files.UploadFiles` instance to read the archive from.
            cache:
                A boolean that indicates if the archive file should be left unclosed,
                e.g. if this method is called for many entries of the same upload.
        '''
        archive = upload_files.read_archive(self.calc_id)
        try:
            # instead of loading the whole archive, it should be enough to load the
            # parts that are referenced by section_metadata/EntryMetadata
            # TODO somehow it should determine which root setions too load from the metainfo
            # or configuration
            calc_archive = archive[self.calc_id]
            entry_archive_dict = {section_metadata: calc_archive[section_metadata].to_dict()}
            if section_workflow in calc_archive:
                entry_archive_dict[section_workflow] = calc_archive[section_workflow].to_dict()
            entry_metadata = datamodel.EntryArchive.m_from_dict(entry_archive_dict)[section_metadata]

        except KeyError:
            # Due hard processing failures, it might be possible that an entry might not
            # have an archive
            if self._entry_metadata is not None:
                entry_metadata = self._entry_metadata

            else:
                entry_metadata = self.create_metadata()

        entry_metadata.m_update_from_dict(self.metadata)

        return entry_metadata

    def user_metadata(self) -> datamodel.EntryMetadata:
        '''
        Returns a :class:`nomad.datamodel.EntryMetadata` with values from this
        processing object and the user metadata, not necessarily the metadata from
        the archive.
        '''
        entry_metadata = self.create_metadata()
        entry_metadata.m_update_from_dict(self.metadata)

        return entry_metadata

    @property
    def upload_files(self) -> ArchiveBasedStagingUploadFiles:
        if not self._upload_files:
            self._upload_files = ArchiveBasedStagingUploadFiles(
                self.upload_id, is_authorized=lambda: True, upload_path=self.upload.upload_path)
        return self._upload_files

    def get_logger(self, **kwargs):
        '''
        Returns a wrapped logger that additionally saves all entries to the calculation
        processing log in the archive.
        '''
        logger = super().get_logger()
        logger = logger.bind(
            upload_id=self.upload_id, mainfile=self.mainfile, calc_id=self.calc_id, **kwargs)

        if self._calc_proc_logs is None:
            self._calc_proc_logs = []

        def save_to_calc_log(logger, method_name, event_dict):
            try:
                # sanitize the event_dict, because all kinds of values might have been added
                dump_dict = {key: str(value) for key, value in event_dict.items()}
                dump_dict.update(level=method_name.upper())
                self._calc_proc_logs.append(dump_dict)

                if method_name == 'error':
                    error = event_dict.get('event', None)
                    if error is not None:
                        self._entry_metadata.processing_errors.append(error)

            except Exception:
                # Exceptions here will cause indefinite loop
                pass

            return event_dict

        return wrap_logger(logger, processors=_log_processors + [save_to_calc_log])

    @process
    def re_process_calc(self):
        '''
        Processes a calculation again. This means there is already metadata and
        instead of creating it initially, we are just updating the existing
        records.
        '''
        logger = self.get_logger()

        if config.reprocess_rematch:
            with utils.timer(logger, 'parser matching executed'):
                parser = match_parser(
                    self.upload_files.raw_file_object(self.mainfile).os_path, strict=False)

            if parser is None and not config.reprocess_unmatched:
                self.errors = ['no parser matches during re-process, will not re-process this calc']

                try:
                    upload_files = PublicUploadFiles(self.upload_id, is_authorized=lambda: True)
                    with upload_files.read_archive(self.calc_id) as archive:
                        self.upload_files.write_archive(self.calc_id, archive[self.calc_id].to_dict())

                except Exception as e:
                    logger.error('could not copy archive for non matching, non reprocessed entry', exc_info=e)
                    raise e

                # mock the steps of actual processing
                self._continue_with('parsing')
                self._continue_with('normalizing')
                self._continue_with('archiving')
                self._complete()
                return

        else:
            parser = parser_dict.get(self.parser)

        if parser is None:
            self.get_logger().warn('no parser matches during re-process, use the old parser')
            self.warnings = ['no matching parser found during re-processing']

        elif self.parser != parser.name:
            if parser_dict[self.parser].name == parser.name:
                # parser was just renamed
                self.parser = parser.name

            else:
                self.parser = parser.name
                logger.info(
                    'different parser matches during re-process, use new parser',
                    parser=parser.name)

        try:
            self._entry_metadata = self.user_metadata()
            self._entry_metadata.calc_hash = self.upload_files.calc_hash(self.mainfile)
            self._entry_metadata.nomad_version = config.meta.version
            self._entry_metadata.nomad_commit = config.meta.commit
            self._entry_metadata.last_processing = datetime.utcnow()
            self._entry_metadata.processing_errors = []
            self._entry_metadata.files = self.upload_files.calc_files(self.mainfile)

            self.parsing()
            self.normalizing()
            self.archiving()
        finally:
            # close loghandler that was not closed due to failures
            try:
                if self._parser_results and self._parser_results.m_resource:
                    self._parser_results.section_metadata = None
                    self._parser_results.m_resource.unload()
            except Exception as e:
                logger.error('could not unload processing results', exc_info=e)

    def _setup_fallback_metadata(self):
        self._entry_metadata = self.create_metadata()
        self._entry_metadata.calc_hash = self.upload_files.calc_hash(self.mainfile)
        self._entry_metadata.last_processing = datetime.utcnow()
        self._entry_metadata.processing_errors = []
        self._entry_metadata.files = self.upload_files.calc_files(self.mainfile)
        self._entry_metadata.parser_name = self.parser

    @process
    def process_calc(self):
        '''
        Processes a new calculation that has no prior records in the mongo, elastic,
        or filesystem storage. It will create an initial set of (user) metadata.
        '''
        logger = self.get_logger()
        if self.upload is None:
            logger.error('calculation upload does not exist')

        try:
            # save preliminary minimum calc metadata in case processing fails
            # successful processing will replace it with the actual metadata
            self._setup_fallback_metadata()

            if len(self._entry_metadata.files) >= config.auxfile_cutoff:
                self.warning(
                    'This calc has many aux files in its directory. '
                    'Have you placed many calculations in the same directory?')

            self.parsing()
            self.normalizing()
            self.archiving()
        finally:
            # close loghandler that was not closed due to failures
            try:
                if self._parser_results and self._parser_results.m_resource:
                    self._parser_results.section_metadata = None
                    self._parser_results.m_resource.unload()
            except Exception as e:
                logger.error('could unload processing results', exc_info=e)

    def on_fail(self):
        # in case of failure, index a minimum set of metadata and mark
        # processing failure
        try:
            if self._entry_metadata is None:
                self._setup_fallback_metadata()

            self._entry_metadata.processed = False

            try:
                self.apply_entry_metadata(self._entry_metadata)
            except Exception as e:
                self.get_logger().error(
                    'could not apply entry metadata to entry', exc_info=e)

            try:
                self._entry_metadata.apply_domain_metadata(self._parser_results)
            except Exception as e:
                self.get_logger().error(
                    'could not apply domain metadata to entry', exc_info=e)

            self._entry_metadata.a_elastic.index()
        except Exception as e:
            self.get_logger().error(
                'could not index after processing failure', exc_info=e)

        try:
            self.write_archive(self._parser_results)
        except Exception as e:
            self.get_logger().error(
                'could not write archive after processing failure', exc_info=e)

    def on_process_complete(self, process_name):
        # the save might be necessary to correctly read the join condition from the db
        self.save()
        # in case of error, the process_name might be unknown
        if process_name == 'process_calc' or process_name == 're_process_calc' or process_name is None:
            self.upload.reload()
            self.upload.check_join()

    @task
    def parsing(self):
        ''' The *task* that encapsulates all parsing related actions. '''
        context = dict(parser=self.parser, step=self.parser)
        logger = self.get_logger(**context)
        parser = parser_dict[self.parser]
        self._entry_metadata.parser_name = self.parser

        with utils.timer(logger, 'parser executed', input_size=self.mainfile_file.size):
            if not config.process_reuse_parser:
                if isinstance(parser, parsing.FairdiParser):
                    try:
                        parser = parser.__class__()
                    except Exception as e:
                        self.fail(
                            'could not re-create parser instance',
                            exc_info=e, error=str(e), **context)
                        return
            try:
                self._parser_results = EntryArchive()
                # allow parsers to read/write metadata
                self._parser_results.m_add_sub_section(EntryArchive.section_metadata, self._entry_metadata)
                parser.parse(
                    self.upload_files.raw_file_object(self.mainfile).os_path,
                    self._parser_results, logger=logger)

            except Exception as e:
                self.fail('parser failed with exception', exc_info=e, error=str(e), **context)
                return
            except SystemExit:
                self.fail('parser raised system exit', error='system exit', **context)
                return

    def process_phonon(self):
        """Function that is run for phonon calculation before cleanup.
        This task is run by the celery process that is calling the join for the
        upload.

        This function re-opens the Archive for this calculation to add method
        information from another referenced archive. Updates the method
        information in section_encyclopedia as well as the DFT domain metadata.
        """
        try:
            logger = self.get_logger(parser=self.parser, step=self.parser)

            # Open the archive of the phonon calculation.
            upload_files = StagingUploadFiles(self.upload_id, is_authorized=lambda: True)
            with upload_files.read_archive(self.calc_id) as archive:
                arch = archive[self.calc_id]
                phonon_archive = EntryArchive.m_from_dict(arch.to_dict())
            self._entry_metadata = phonon_archive.section_metadata
            self._calc_proc_logs = phonon_archive.processing_logs

            # Re-create the parse results
            self._parser_results = phonon_archive

            # Read in the first referenced calculation. The reference is given as
            # an absolute path which needs to be converted into a path that is
            # relative to upload root.
            scc = self._parser_results.section_run[0].section_single_configuration_calculation[0]
            calculation_refs = scc.section_calculation_to_calculation_refs
            if calculation_refs is None:
                logger.error("No calculation_to_calculation references found")
                return

            relative_ref = scc.section_calculation_to_calculation_refs[0].calculation_to_calculation_external_url
            ref_id = upload_files.calc_id(relative_ref)
            with upload_files.read_archive(ref_id) as archive:
                arch = archive[ref_id]
                ref_archive = EntryArchive.m_from_dict(arch.to_dict())

            # Get encyclopedia method information directly from the referenced calculation.
            ref_enc_method = ref_archive.section_metadata.encyclopedia.method
            if ref_enc_method is None or len(ref_enc_method) == 0 or ref_enc_method.functional_type is None:
                logger.error("No method information available in referenced calculation.")
                return

            self._parser_results.section_metadata.encyclopedia.method = ref_enc_method

            # Overwrite old entry with new data. The metadata is updated with
            # new timestamp and method details taken from the referenced
            # archive.
            self._entry_metadata.last_processing = datetime.utcnow()
            self._entry_metadata.dft.xc_functional = ref_archive.section_metadata.dft.xc_functional
            self._entry_metadata.dft.basis_set = ref_archive.section_metadata.dft.basis_set
            self._entry_metadata.dft.update_group_hash()
            self._entry_metadata.encyclopedia.status = EncyclopediaMetadata.status.type.success
        except Exception as e:
            logger.error("Could not retrieve method information for phonon calculation.", exc_info=e)
            if self._entry_metadata is None:
                self._setup_fallback_metadata()
                self._entry_metadata.processed = False

            try:
                if self._entry_metadata.encyclopedia is None:
                    self._entry_metadata.encyclopedia = EncyclopediaMetadata()
                self._entry_metadata.encyclopedia.status = EncyclopediaMetadata.status.type.failure
            except Exception as e:
                logger.error("Could set encyclopedia status.", exc_info=e)

        finally:
            # persist the calc metadata
            with utils.timer(logger, 'calc metadata saved'):
                self.apply_entry_metadata(self._entry_metadata)

            # index in search
            with utils.timer(logger, 'calc metadata indexed'):
                self._entry_metadata.a_elastic.index()

            # persist the archive
            with utils.timer(
                    logger, 'calc archived',
                    input_size=self.mainfile_file.size) as log_data:

                archive_size = self.write_archive(self._parser_results)
                log_data.update(archive_size=archive_size)

    @task
    def normalizing(self):
        ''' The *task* that encapsulates all normalizing related actions. '''

        # allow normalizer to access and add data to the entry metadata
        if self._parser_results.section_metadata is None:
            self._parser_results.m_add_sub_section(
                datamodel.EntryArchive.section_metadata, self._entry_metadata)

        for normalizer in normalizers:
            if normalizer.domain != parser_dict[self.parser].domain:
                continue

            normalizer_name = normalizer.__name__
            context = dict(normalizer=normalizer_name, step=normalizer_name)
            logger = self.get_logger(**context)

            with utils.timer(logger, 'normalizer executed', input_size=self.mainfile_file.size):
                try:
                    normalizer(self._parser_results).normalize(logger=logger)
                    logger.info('normalizer completed successfull', **context)
                except Exception as e:
                    self.fail('normalizer failed with exception', exc_info=e, error=str(e), **context)

    def _read_metadata_from_file(self, logger):
        # metadata file name defined in nomad.config nomad_metadata.yaml/json
        # which can be placed in the directory containing the mainfile or somewhere up
        # highest priority is directory with mainfile
        metadata_file = config.metadata_file_name
        metadata_dir = os.path.dirname(self.mainfile_file.os_path)
        upload_raw_dir = self.upload_files._raw_dir.os_path

        metadata = {}
        metadata_part = None
        # apply the nomad files of the current directory and parent directories
        while True:
            metadata_part = self.upload.metadata_file_cached(
                os.path.join(metadata_dir, metadata_file))
            for key, val in metadata_part.items():
                if key in ['entries', 'oasis_datasets']:
                    continue
                metadata.setdefault(key, val)

            if metadata_dir == upload_raw_dir:
                break

            metadata_dir = os.path.dirname(metadata_dir)

        # Top-level nomad file can also contain an entries dict with entry
        # metadata per mainfile as key. This takes precedence of the other files.
        entries = metadata_part.get('entries', {})
        metadata_part = entries.get(self.mainfile, {})
        for key, val in metadata_part.items():
            metadata[key] = val

        if len(metadata) > 0:
            logger.info('Apply user metadata from nomad.yaml/json file')

        for key, val in metadata.items():
            if key in ['entries', 'skip_matching']:
                continue

            definition = _editable_metadata.get(key, None)
            if definition is None and self.upload.from_oasis:
                definition = _oasis_metadata.get(key, None)

            if key == 'uploader':
                if datamodel.User.get(self.upload.user_id).is_admin:
                    definition = datamodel.EntryMetadata.uploader

            if definition is None:
                logger.warn('Users cannot set metadata', quantity=key)
                continue

            try:
                self._entry_metadata.m_set(definition, val)
                if definition == datamodel.EntryMetadata.calc_id:
                    self.calc_id = val
            except Exception as e:
                logger.error(
                    'Could not apply user metadata from nomad.yaml/json file',
                    quantitiy=definition.name, exc_info=e)

    @task
    def archiving(self):
        ''' The *task* that encapsulates all archival related actions. '''
        logger = self.get_logger()

        self._entry_metadata.apply_domain_metadata(self._parser_results)
        self._entry_metadata.processed = True

        if self.upload.publish_directly:
            self._entry_metadata.published = False

        try:
            self._read_metadata_from_file(logger)
        except Exception as e:
            logger.error('could not process user metadata in nomad.yaml/json file', exc_info=e)

        # persist the calc metadata
        with utils.timer(logger, 'calc metadata saved'):
            self.apply_entry_metadata(self._entry_metadata)

        # index in search
        with utils.timer(logger, 'calc metadata indexed'):
            self._entry_metadata.a_elastic.index()

        # persist the archive
        with utils.timer(
                logger, 'calc archived',
                input_size=self.mainfile_file.size) as log_data:

            archive_size = self.write_archive(self._parser_results)
            log_data.update(archive_size=archive_size)

    def write_archive(self, archive: EntryArchive):
        # save the archive mongo entry
        try:
            if self._entry_metadata.processed:
                write_partial_archive_to_mongo(archive)
        except Exception as e:
            self.get_logger().error('could not write mongodb archive entry', exc_info=e)

        # add the processing logs to the archive
        def filter_processing_logs(logs):
            if len(logs) > 100:
                return [
                    log for log in logs
                    if log.get('level') != 'DEBUG']
            return logs

        if self._calc_proc_logs is None:
            self._calc_proc_logs = []

        if archive is not None:
            archive = archive.m_copy()
        else:
            archive = datamodel.EntryArchive()

        if archive.section_metadata is None:
            archive.m_add_sub_section(datamodel.EntryArchive.section_metadata, self._entry_metadata)

        archive.processing_logs = filter_processing_logs(self._calc_proc_logs)

        # save the archive msg-pack
        try:
            return self.upload_files.write_archive(self.calc_id, archive.m_to_dict())
        except Exception as e:
            # most likely failed due to domain data, try to write metadata and processing logs
            archive = datamodel.EntryArchive()
            archive.m_add_sub_section(datamodel.EntryArchive.section_metadata, self._entry_metadata)
            archive.processing_logs = filter_processing_logs(self._calc_proc_logs)
            self.upload_files.write_archive(self.calc_id, archive.m_to_dict())
            raise e

    def __str__(self):
        return 'calc %s calc_id=%s upload_id%s' % (super().__str__(), self.calc_id, self.upload_id)


class Upload(Proc):
    '''
    Represents uploads in the databases. Provides persistence access to the files storage,
    and processing state.

    Attributes:
        name: Optional user provided upload name.
        upload_path: The fs path were the uploaded files was stored during upload.
        temporary: True if the uploaded file should be removed after extraction.

        upload_id: The upload id generated by the database or the uploaded NOMAD deployment.
        upload_time: Datetime of the original upload independent of the NOMAD deployment
            it was first uploaded to.
        user_id: The id of the user that created this upload.
        published: Boolean that indicates that the upload is published on this NOMAD deployment.
        publish_time: Datetime when the upload was initially published on this NOMAD deployment.
        last_update: Datetime of the last modifying process run (publish, re-processing, upload).

        publish_directly: Boolean indicating that this upload should be published after initial processing.
        from_oasis: Boolean indicating that this upload is coming from another NOMAD deployment.
        oasis_id: The deployment id of the NOMAD that uploaded the upload.
        published_to: A list of deployment ids where this upload has been successfully uploaded to.

        joined: Boolean indicates if the running processing has joined (:func:`check_join`).
    '''
    id_field = 'upload_id'

    upload_id = StringField(primary_key=True)
    upload_path = StringField(default=None)
    temporary = BooleanField(default=False)
    embargo_length = IntField(default=36)

    name = StringField(default=None)
    upload_time = DateTimeField()
    user_id = StringField(required=True)
    published = BooleanField(default=False)
    publish_time = DateTimeField()
    last_update = DateTimeField()

    publish_directly = BooleanField(default=False)
    from_oasis = BooleanField(default=False)
    oasis_deployment_id = StringField(default=None)
    published_to = ListField(StringField())

    joined = BooleanField(default=False)

    meta: Any = {
        'strict': False,
        'indexes': [
            'user_id', 'tasks_status', 'process_status', 'published', 'upload_time', 'create_time'
        ]
    }

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.publish_directly = self.publish_directly or self.from_oasis
        self._upload_files: ArchiveBasedStagingUploadFiles = None

    @lru_cache()
    def metadata_file_cached(self, path):
        for ext in config.metadata_file_extensions:
            full_path = '%s.%s' % (path, ext)
            if os.path.isfile(full_path):
                try:
                    with open(full_path) as f:
                        if full_path.endswith('.json'):
                            return json.load(f)
                        elif full_path.endswith('.yaml') or full_path.endswith('.yml'):
                            return yaml.load(f, Loader=getattr(yaml, 'FullLoader'))
                        else:
                            return {}
                except Exception as e:
                    self.get_logger().warn('could not parse nomad.yaml/json', path=path, exc_info=e)
                    # ignore the file contents if the file is not parsable
                    pass
        return {}

    @property
    def metadata(self) -> dict:
        '''
        Getter, setter for user metadata. Metadata is pickled to and from the public
        bucket to allow sharing among all processes. Usually uploads do not have (much)
        user defined metadata, but users provide all metadata per upload as part of
        the publish process. This will change, when we introduce editing functionality
        and metadata will be provided through different means.
        '''
        try:
            upload_files = PublicUploadFiles(self.upload_id, is_authorized=lambda: True)
        except KeyError:
            return None
        return upload_files.user_metadata

    @metadata.setter
    def metadata(self, data: dict) -> None:
        upload_files = PublicUploadFiles(self.upload_id, is_authorized=lambda: True, create=True)
        upload_files.user_metadata = data

    @classmethod
    def get(cls, id: str, include_published: bool = True) -> 'Upload':
        return cls.get_by_id(id, 'upload_id')

    @classmethod
    def user_uploads(cls, user: datamodel.User, **kwargs) -> List['Upload']:
        ''' Returns all uploads for the given user. Kwargs are passed to mongo query. '''
        return cls.objects(user_id=str(user.user_id), **kwargs)

    @property
    def uploader(self):
        return datamodel.User.get(self.user_id)

    def get_logger(self, **kwargs):
        logger = super().get_logger()
        user = self.uploader
        user_name = '%s %s' % (user.first_name, user.last_name)
        # We are not using 'user_id' because logstash (?) will filter these entries ?!
        logger = logger.bind(
            upload_id=self.upload_id, upload_name=self.name, user_name=user_name,
            user=user.user_id, **kwargs)
        return logger

    @classmethod
    def create(cls, **kwargs) -> 'Upload':
        '''
        Creates a new upload for the given user, a user given name is optional.
        It will populate the record with a signed url and pending :class:`UploadProc`.
        The upload will be already saved to the database.

        Arguments:
            user: The user that created the upload.
        '''
        # use kwargs to keep compatibility with super method
        user: datamodel.User = kwargs['user']
        del(kwargs['user'])

        if 'upload_id' not in kwargs:
            kwargs.update(upload_id=utils.create_uuid())
        kwargs.update(user_id=user.user_id)
        self = super().create(**kwargs)

        self._continue_with('uploading')

        return self

    def delete(self):
        ''' Deletes this upload process state entry and its calcs. '''
        Calc.objects(upload_id=self.upload_id).delete()
        super().delete()

    def delete_upload_local(self):
        '''
        Deletes the upload, including its processing state and
        staging files. Local version without celery processing.
        '''
        logger = self.get_logger(upload_size=self.upload_files.size)

        with utils.lnr(logger, 'upload delete failed'):
            with utils.timer(logger, 'upload deleted from index'):
                search.delete_upload(self.upload_id)

            with utils.timer(logger, 'upload partial archives deleted'):
                calc_ids = [calc.calc_id for calc in Calc.objects(upload_id=self.upload_id)]
                delete_partial_archives_from_mongo(calc_ids)

            with utils.timer(logger, 'upload files deleted'):
                self.upload_files.delete()

            self.delete()

    @process
    def delete_upload(self):
        '''
        Deletes the upload, including its processing state and
        staging files. This starts the celery process of deleting the upload.
        '''
        self.delete_upload_local()

        return True  # do not save the process status on the delete upload

    @process
    def publish_upload(self):
        '''
        Moves the upload out of staging to the public area. It will
        pack the staging upload files in to public upload files.
        '''
        assert self.processed_calcs > 0

        logger = self.get_logger(upload_size=self.upload_files.size)
        logger.info('started to publish')

        with utils.lnr(logger, 'publish failed'):
            with self.entries_metadata(self.metadata) as calcs:

                with utils.timer(logger, 'upload metadata updated'):
                    def create_update(calc):
                        calc.published = True
                        calc.with_embargo = calc.with_embargo if calc.with_embargo is not None else False
                        return UpdateOne(
                            {'_id': calc.calc_id},
                            {'$set': {'metadata': calc.m_to_dict(
                                include_defaults=True, categories=[datamodel.MongoMetadata])}})

                    Calc._get_collection().bulk_write([create_update(calc) for calc in calcs])

                if isinstance(self.upload_files, StagingUploadFiles):
                    with utils.timer(logger, 'staged upload files packed'):
                        self.upload_files.pack(calcs)

                with utils.timer(logger, 'index updated'):
                    search.publish(calcs)

                if isinstance(self.upload_files, StagingUploadFiles):
                    with utils.timer(logger, 'upload staging files deleted'):
                        self.upload_files.delete()
                        self.published = True
                        self.publish_time = datetime.utcnow()
                        self.last_update = datetime.utcnow()
                        self.save()
                else:
                    self.last_update = datetime.utcnow()
                    self.save()

    @process
    def publish_from_oasis(self):
        '''
        Uploads the already published upload to a different NOMAD deployment. This allows
        to push uploads from an OASIS to the central NOMAD.
        '''
        assert self.published, \
            'Only published uploads can be published to the central NOMAD.'
        assert config.oasis.central_nomad_deployment_id not in self.published_to, \
            'Upload is already published to the central NOMAD.'

        from nomad.cli.client.client import _create_client as create_client
        central_nomad_client = create_client(
            user=config.keycloak.username,
            password=config.keycloak.password,
            api_base_url=config.oasis.central_nomad_api_url,
            use_token=False)

        # compile oasis metadata for the upload
        upload_metadata = dict(upload_time=str(self.upload_time))
        upload_metadata_entries = {}
        upload_metadata_datasets = {}
        for calc in self.calcs:
            entry_metadata = dict(**{
                key: str(value) if isinstance(value, datetime) else value
                for key, value in calc.metadata.items()
                if key in _editable_metadata or key in _oasis_metadata})
            entry_metadata['calc_id'] = calc.calc_id
            if entry_metadata.get('with_embargo'):
                continue
            upload_metadata_entries[calc.mainfile] = entry_metadata
            if 'datasets' in entry_metadata:
                for dataset_id in entry_metadata['datasets']:
                    if dataset_id in upload_metadata_datasets:
                        continue

                    dataset = datamodel.Dataset.m_def.a_mongo.get(dataset_id=dataset_id)
                    upload_metadata_datasets[dataset_id] = dataset.m_to_dict()

        upload_metadata['entries'] = upload_metadata_entries
        upload_metadata['oasis_datasets'] = {
            dataset['name']: dataset for dataset in upload_metadata_datasets.values()}
        oasis_upload_id, upload_metadata = _normalize_oasis_upload_metadata(
            self.upload_id, upload_metadata)

        self.last_status_message = 'Compiled metadata to upload to the central NOMAD.'
        self.save()

        assert len(upload_metadata_entries) > 0, \
            'Only uploads with public contents can be published to the central NOMAD.'

        # add oasis metadata to the upload
        public_upload_files = cast(PublicUploadFiles, self.upload_files)
        public_upload_files.add_metadata_file(upload_metadata)
        file_to_upload = public_upload_files.public_raw_data_file

        self.last_status_message = 'Prepared the upload for uploading to central NOMAD.'
        self.save()

        # upload to central NOMAD
        oasis_admin_token = central_nomad_client.auth.get_auth().response().result.access_token
        upload_headers = dict(Authorization='Bearer %s' % oasis_admin_token)
        upload_parameters = dict(
            oasis_upload_id=oasis_upload_id,
            oasis_uploader_id=self.user_id,
            oasis_deployment_id=config.meta.deployment_id)
        upload_url = '%s/uploads/?%s' % (
            config.oasis.central_nomad_api_url,
            urllib.parse.urlencode(upload_parameters))

        with open(file_to_upload, 'rb') as f:
            response = requests.put(upload_url, headers=upload_headers, data=f)

        if response.status_code != 200:
            self.get_logger().error(
                'Could not upload to central NOMAD', status_code=response.status_code)
            self.last_status_message = 'Could not upload to central NOMAD.'
            return

        self.published_to.append(config.oasis.central_nomad_deployment_id)
        self.last_status_message = 'Successfully uploaded to central NOMAD.'

    @process
    def re_process_upload(self):
        '''
        A *process* that performs the re-processing of a earlier processed
        upload.

        Runs the distributed process of fully reparsing/re-normalizing an existing and
        already published upload. Will renew the archive part of the upload and update
        mongo and elastic search entries.

        TODO this implementation does not do any re-matching. This will be more complex
        due to handling of new or missing matches.
        '''
        logger = self.get_logger()
        logger.info('started to re-process')

        # mock the steps of actual processing
        self._continue_with('uploading')

        # extract the published raw files into a staging upload files instance
        self._continue_with('extracting')

        if self.published:
            try:
                staging_upload_files = StagingUploadFiles(self.upload_id)
                # public files exist and there is a staging directory, it is probably old
                # and we delete it first
                with utils.timer(logger, 'upload staging files deleted'):
                    staging_upload_files.delete()
                    logger.warn('deleted old staging files')

            except KeyError as e:
                logger.info('reprocessing published files')
        else:
            logger.info('reprocessing staging files')

        with utils.timer(logger, 'upload extracted'):
            staging_upload_files = self.upload_files.to_staging_upload_files(create=True)

        self._continue_with('parse_all')
        try:
            if config.reprocess_match:
                with utils.timer(logger, 'calcs match on reprocess'):
                    with_embargo_values = Calc._get_collection().distinct(
                        'metadata.with_embargo', dict(upload_id=self.upload_id))
                    with_embargo = with_embargo_values != [False]

                    for filename, parser in self.match_mainfiles():
                        calc_id = staging_upload_files.calc_id(filename)
                        try:
                            Calc.get(calc_id)
                        except KeyError:
                            calc = Calc.create(
                                calc_id=calc_id,
                                mainfile=filename, parser=parser.name,
                                worker_hostname=self.worker_hostname,
                                create_time=self.upload_time,
                                upload_id=self.upload_id)

                            calc.metadata.update(
                                published=self.published,
                                with_embargo=with_embargo)

                            calc.save()

            with utils.timer(logger, 'calcs resetted'):
                # check if a calc is already/still processing
                processing = Calc.objects(
                    upload_id=self.upload_id,
                    **Calc.process_running_mongoengine_query()).count()

                if processing > 0:
                    logger.warn(
                        'processes are still/already running on calc, they will be resetted',
                        count=processing)

                # reset all calcs
                Calc._get_collection().update_many(
                    dict(upload_id=self.upload_id),
                    {'$set': Calc.reset_pymongo_update(worker_hostname=self.worker_hostname)})

            with utils.timer(logger, 'calcs re-processing called'):
                # process call calcs
                Calc.process_all(Calc.re_process_calc, dict(upload_id=self.upload_id), exclude=['metadata'])

                logger.info('completed to trigger re-process of all calcs')

        except Exception as e:
            # try to remove the staging copy in failure case
            logger.error('failed to trigger re-process of all calcs', exc_info=e)

            if self.published:
                if staging_upload_files is not None and staging_upload_files.exists():
                    staging_upload_files.delete()

            raise e

        # the packing and removing of the staging upload files, will be trigged by
        # the 'cleanup' task after processing all calcs

    @process
    def re_pack(self):
        ''' A *process* that repacks the raw and archive data based on the current embargo data. '''
        assert self.published
        self.upload_files.re_pack(self.user_metadata())

    @process
    def process_upload(self):
        ''' A *process* that performs the initial upload processing. '''
        self.extracting()

        if self.from_oasis:
            # we might need to add datasets from the oasis before processing and
            # adding the entries
            oasis_metadata_file = os.path.join(self.upload_files._raw_dir.os_path, config.metadata_file_name + '.json')
            with open(oasis_metadata_file, 'rt') as f:
                oasis_metadata = json.load(f)
            oasis_datasets = oasis_metadata.get('oasis_datasets', {})
            metadata_was_changed = False
            for oasis_dataset in oasis_datasets.values():
                try:
                    existing_dataset = datamodel.Dataset.m_def.a_mongo.get(
                        user_id=self.user_id, name=oasis_dataset['name'])
                except KeyError:
                    datamodel.Dataset(**oasis_dataset).a_mongo.save()
                else:
                    oasis_dataset_id = oasis_dataset['dataset_id']
                    if existing_dataset.dataset_id != oasis_dataset_id:
                        # A dataset for the same user with the same name was created
                        # in both deployments. We consider this to be the "same" dataset.
                        # These datasets have different ids and we need to migrate the provided
                        # dataset ids:
                        for entry in oasis_metadata['entries'].values():
                            entry_datasets = entry.get('datasets', [])
                            for index, dataset_id in enumerate(entry_datasets):
                                if dataset_id == oasis_dataset_id:
                                    entry_datasets[index] = existing_dataset.dataset_id
                                    metadata_was_changed = True

            if metadata_was_changed:
                with open(oasis_metadata_file, 'wt') as f:
                    json.dump(oasis_metadata, f)

        self.parse_all()

    @task
    def uploading(self):
        ''' A no-op *task* as a stand-in for receiving upload data. '''
        pass

    @property
    def upload_files(self) -> UploadFiles:
        upload_files_class = ArchiveBasedStagingUploadFiles if not self.published else PublicUploadFiles
        kwargs = dict(upload_path=self.upload_path) if not self.published else {}

        if not self._upload_files or not isinstance(self._upload_files, upload_files_class):
            self._upload_files = upload_files_class(
                self.upload_id, is_authorized=lambda: True, **kwargs)

        return self._upload_files

    @property
    def staging_upload_files(self) -> StagingUploadFiles:
        return self.upload_files.to_staging_upload_files()

    @task
    def extracting(self):
        '''
        The *task* performed before the actual parsing/normalizing: extracting
        the uploaded files.
        '''
        # extract the uploaded file
        self._upload_files = ArchiveBasedStagingUploadFiles(
            upload_id=self.upload_id, is_authorized=lambda: True, create=True,
            upload_path=self.upload_path)

        logger = self.get_logger()
        try:
            with utils.timer(logger, 'upload extracted', upload_size=self.upload_files.size):
                self.upload_files.extract()

            if self.temporary:
                os.remove(self.upload_path)
                self.upload_path = None

        except KeyError:
            self.fail('processing requested for non existing upload', log_level=logging.ERROR)
            return
        except ExtractError:
            self.fail('bad .zip/.tar file', log_level=logging.INFO)
            return

    def _preprocess_files(self, path):
        '''
        Some files need preprocessing. Currently we need to add a stripped POTCAR version
        and always restrict/embargo the original.
        '''
        if os.path.basename(path).startswith('POTCAR'):
            # create checksum
            hash = hashlib.sha224()
            with open(self.staging_upload_files.raw_file_object(path).os_path, 'rb') as orig_f:
                for line in orig_f.readlines():
                    hash.update(line)

            checksum = hash.hexdigest()

            # created stripped POTCAR
            stripped_path = path + '.stripped'
            with open(self.staging_upload_files.raw_file_object(stripped_path).os_path, 'wt') as stripped_f:
                stripped_f.write('Stripped POTCAR file. Checksum of original file (sha224): %s\n' % checksum)
            os.system(
                '''
                    awk < '%s' >> '%s' '
                    BEGIN { dump=1 }
                    /End of Dataset/ { dump=1 }
                    dump==1 { print }
                    /END of PSCTR/ { dump=0 }'
                ''' % (
                    self.staging_upload_files.raw_file_object(path).os_path,
                    self.staging_upload_files.raw_file_object(stripped_path).os_path))

    def match_mainfiles(self) -> Iterator[Tuple[str, object]]:
        '''
        Generator function that matches all files in the upload to all parsers to
        determine the upload's mainfiles.

        Returns:
            Tuples of mainfile, filename, and parsers
        '''
        metadata = self.metadata_file_cached(
            os.path.join(self.upload_files.os_path, 'raw', config.metadata_file_name))
        skip_matching = metadata.get('skip_matching', False)
        entries_metadata = metadata.get('entries', {})

        directories_with_match: Dict[str, str] = dict()
        upload_files = self.upload_files.to_staging_upload_files()
        for filename in upload_files.raw_file_manifest():
            self._preprocess_files(filename)

            if skip_matching and filename not in entries_metadata:
                continue

            try:
                parser = match_parser(upload_files.raw_file_object(filename).os_path)
                if parser is not None:
                    directory = os.path.dirname(filename)
                    if directory in directories_with_match:
                        # TODO this might give us the chance to store directory based relationship
                        # between calcs for the future?
                        pass
                    else:
                        directories_with_match[directory] = filename

                    yield filename, parser
            except Exception as e:
                self.get_logger().error(
                    'exception while matching pot. mainfile',
                    mainfile=filename, exc_info=e)

    @task
    def parse_all(self):
        '''
        The *task* used to identify mainfile/parser combinations among the upload's files, creates
        respective :class:`Calc` instances, and triggers their processing.
        '''
        logger = self.get_logger()

        oasis_metadata = {}
        if self.from_oasis:
            oasis_metadata = self.metadata_file_cached(
                os.path.join(self.upload_files.os_path, 'raw', config.metadata_file_name)).get('entries', {})

        with utils.timer(logger, 'calcs processing called'):
            for filename, parser in self.match_mainfiles():
                oasis_entry_metadata = oasis_metadata.get(filename)
                if oasis_entry_metadata is not None:
                    calc_id = oasis_entry_metadata.get('calc_id')
                    if calc_id is None:
                        logger.warn('Oasis entry without id', mainfile=filename)
                        calc_id = self.upload_files.calc_id(filename)
                else:
                    calc_id = self.upload_files.calc_id(filename)

                calc = Calc.create(
                    calc_id=calc_id,
                    mainfile=filename, parser=parser.name,
                    worker_hostname=self.worker_hostname,
                    upload_id=self.upload_id)

                calc.process_calc()

    def on_process_complete(self, process_name):
        if process_name == 'process_upload' or process_name == 're_process_upload':
            self.check_join()

    def check_join(self):
        '''
        Performs an evaluation of the join condition and triggers the :func:`cleanup`
        task if necessary. The join condition allows to run the ``cleanup`` after
        all calculations have been processed. The upload processing stops after all
        calculation processing have been triggered (:func:`parse_all` or
        :func:`re_process_upload`). The cleanup task is then run within the last
        calculation process (the one that triggered the join by calling this method).
        '''
        total_calcs = self.total_calcs
        processed_calcs = self.processed_calcs

        self.get_logger().debug('check join', processed_calcs=processed_calcs, total_calcs=total_calcs)
        # check if process is not running anymore, i.e. not still spawining new processes to join
        # check the join condition, i.e. all calcs have been processed
        if not self.process_running and processed_calcs >= total_calcs:
            # this can easily be called multiple times, e.g. upload finished after all calcs finished
            modified_upload = self._get_collection().find_one_and_update(
                {'_id': self.upload_id, 'joined': {'$ne': True}},
                {'$set': {'joined': True}})
            if modified_upload is None or modified_upload['joined'] is False:
                self.get_logger().info('join')

                # Before cleaning up, run an additional normalizer on phonon
                # calculations. TODO: This should be replaced by a more
                # extensive mechanism that supports more complex dependencies
                # between calculations.
                phonon_calculations = Calc.objects(upload_id=self.upload_id, parser="parsers/phonopy")
                for calc in phonon_calculations:
                    calc.process_phonon()

                self.cleanup()
            else:
                # the join was already done due to a prior call
                pass

    def reset(self, force=False):
        self.joined = False
        super().reset(force=force)

    @classmethod
    def reset_pymongo_update(cls, worker_hostname: str = None):
        update = super().reset_pymongo_update()
        update.update(joined=False)
        return update

    def _cleanup_after_processing(self):
        # send email about process finish
        if not self.publish_directly:
            user = self.uploader
            name = '%s %s' % (user.first_name, user.last_name)
            message = '\n'.join([
                'Dear %s,' % name,
                '',
                'your data %suploaded at %s has completed processing.' % (
                    '"%s" ' % self.name if self.name else '', self.upload_time.isoformat()),  # pylint: disable=no-member
                'You can review your data on your upload page: %s' % config.gui_url(page='uploads'),
                '',
                'If you encounter any issues with your upload, please let us know and reply to this email.',
                '',
                'The nomad team'
            ])
            try:
                infrastructure.send_mail(
                    name=name, email=user.email, message=message, subject='Processing completed')
            except Exception as e:
                # probably due to email configuration problems
                # don't fail or present this error to clients
                self.logger.error('could not send after processing email', exc_info=e)

        if not self.publish_directly or self.processed_calcs == 0:
            return

        logger = self.get_logger(upload_size=self.upload_files.size)
        logger.info('started to publish upload directly')

        with utils.lnr(logger, 'publish failed'):
            metadata = self.metadata_file_cached(
                os.path.join(self.upload_files.os_path, 'raw', config.metadata_file_name))

            with self.entries_metadata(self.metadata) as calcs:
                with utils.timer(logger, 'upload metadata updated'):
                    def create_update(calc):
                        calc.published = True
                        calc.with_embargo = calc.with_embargo if calc.with_embargo is not None else False
                        return UpdateOne(
                            {'_id': calc.calc_id},
                            {'$set': {'metadata': calc.m_to_dict(
                                include_defaults=True, categories=[datamodel.MongoMetadata])}})

                    Calc._get_collection().bulk_write([create_update(calc) for calc in calcs])

                with utils.timer(logger, 'upload staging files packed'):
                    self.upload_files.pack(calcs)

                with utils.timer(logger, 'index updated'):
                    search.publish(calcs)

            with utils.timer(logger, 'upload staging files deleted'):
                self.upload_files.delete()

            if self.from_oasis:
                if metadata is not None:
                    self.upload_time = metadata.get('upload_time')

                if self.upload_time is None:
                    self.upload_time = datetime.utcnow()
                    logger.warn('oasis upload without upload time')

            self.publish_time = datetime.utcnow()
            self.published = True
            self.last_update = datetime.utcnow()
            self.save()

    def _cleanup_after_re_processing(self):
        logger = self.get_logger(upload_size=self.upload_files.size)
        if self.published:
            staging_upload_files = self.staging_upload_files
            logger.info('started to repack re-processed upload')

            with utils.timer(logger, 'staged upload files re-packed'):
                staging_upload_files.pack(self.user_metadata(), skip_raw=True)

            with utils.timer(logger, 'staged upload files deleted'):
                staging_upload_files.delete()
                self.last_update = datetime.utcnow()
                self.save()

        else:
            logger.info('no cleanup after re-processing unpublished upload')

    @task
    def cleanup(self):
        '''
        The *task* that "cleans" the processing, i.e. removed obsolete files and performs
        pending archival operations. Depends on the type of processing.
        '''
        search.refresh()

        if self.current_process == 're_process_upload':
            self._cleanup_after_re_processing()
        else:
            self._cleanup_after_processing()

    def get_calc(self, calc_id) -> Calc:
        ''' Returns the upload calc with the given id or ``None``. '''
        return Calc.objects(upload_id=self.upload_id, calc_id=calc_id).first()

    @property
    def processed_calcs(self):
        '''
        The number of successfully or not successfully processed calculations. I.e.
        calculations that have finished processing.
        '''
        return Calc.objects(upload_id=self.upload_id, tasks_status__in=[SUCCESS, FAILURE]).count()

    @property
    def total_calcs(self):
        ''' The number of all calculations. '''
        return Calc.objects(upload_id=self.upload_id).count()

    @property
    def failed_calcs(self):
        ''' The number of calculations with failed processing. '''
        return Calc.objects(upload_id=self.upload_id, tasks_status=FAILURE).count()

    @property
    def pending_calcs(self) -> int:
        ''' The number of calculations with pending processing. '''
        return Calc.objects(upload_id=self.upload_id, tasks_status=PENDING).count()

    def all_calcs(self, start, end, order_by=None):
        '''
        Returns all calculations, paginated and ordered.

        Arguments:
            start: the start index of the requested page
            end: the end index of the requested page
            order_by: the property to order by
        '''
        query = Calc.objects(upload_id=self.upload_id)[start:end]
        if not order_by:
            return query
        if type(order_by) == str:
            return query.order_by(order_by)
        assert type(order_by) == tuple, 'order_by must be a string or a tuple if set'
        return query.order_by(*order_by)

    @property
    def outdated_calcs(self):
        ''' All successfully processed and outdated calculations. '''
        return Calc.objects(
            upload_id=self.upload_id, tasks_status=SUCCESS,
            metadata__nomad_version__ne=config.meta.version)

    @property
    def calcs(self):
        ''' All successfully processed calculations. '''
        return Calc.objects(upload_id=self.upload_id, tasks_status=SUCCESS)

    @contextmanager
    def entries_metadata(
            self, user_metadata: dict = None) -> Iterator[Iterable[datamodel.EntryMetadata]]:
        '''
        This is the :py:mod:`nomad.datamodel` transformation method to transform
        processing upload's entries into list of :class:`nomad.datamodel.EntryMetadata` objects.

        Arguments:
            user_metadata: A dict of user metadata that is applied to the resulting
                datamodel data and the respective calculations.
        '''
        upload_files = self.upload_files

        # prepare user metadata per upload and per calc
        if user_metadata is not None:
            entries_metadata_dict: Dict[str, Any] = dict()
            upload_metadata: Dict[str, Any] = dict()

            upload_metadata.update(user_metadata)
            if 'calculations' in upload_metadata:
                del(upload_metadata['calculations'])

            for calc in user_metadata.get('calculations', []):  # pylint: disable=no-member
                entries_metadata_dict[calc['mainfile']] = calc

            user_upload_time = upload_metadata.get('upload_time', None)
            user_upload_name = upload_metadata.get('upload_name', None)

            def get_metadata(calc: Calc):
                entry_metadata = calc.entry_metadata(upload_files)
                entry_user_metadata = dict(upload_metadata)
                entry_user_metadata.pop('embargo_length', None)  # this is for uploads only
                entry_user_metadata.update(entries_metadata_dict.get(calc.mainfile, {}))
                entry_metadata.apply_user_metadata(entry_user_metadata)
                if entry_metadata.upload_time is None:
                    entry_metadata.upload_time = self.upload_time if user_upload_time is None else user_upload_time
                if entry_metadata.upload_name is None:
                    entry_metadata.upload_name = self.name if user_upload_name is None else user_upload_name

                return entry_metadata
        else:
            user_upload_time = None

            def get_metadata(calc: Calc):
                entry_metadata = calc.entry_metadata(upload_files)
                entry_metadata.upload_time = self.upload_time
                entry_metadata.upload_name = self.name

                return entry_metadata

        try:
            # read all calc objects first to avoid missing curser errors
            yield [
                get_metadata(calc)
                for calc in list(Calc.objects(upload_id=self.upload_id))]

        finally:
            upload_files.close()

    def entry_ids(self) -> Iterable[str]:
        return [calc.calc_id for calc in Calc.objects(upload_id=self.upload_id)]

    def user_metadata(self) -> Iterable[datamodel.EntryMetadata]:
        return [calc.user_metadata() for calc in Calc.objects(upload_id=self.upload_id)]

    def compress_and_set_metadata(self, metadata: Dict[str, Any]) -> None:
        '''
        Stores the given user metadata in the upload document. This is the metadata
        adhering to the API model (``UploadMetaData``). Most quantities can be stored
        for the upload and for each calculation. This method will try to move same values
        from the calculation to the upload to "compress" the data.
        '''
        self.embargo_length = min(metadata.get('embargo_length', 36), 36)

        compressed = {
            key: value for key, value in metadata.items() if key != 'calculations'}
        calculations: List[Dict[str, Any]] = []
        compressed['calculations'] = calculations

        for calc in metadata.get('calculations', []):
            compressed_calc: Dict[str, Any] = {}
            calculations.append(compressed_calc)
            for key, value in calc.items():
                if key in ['pid', 'mainfile', 'external_id']:
                    # these quantities are explicitly calc specific and have to stay with
                    # the calc
                    compressed_calc[key] = value
                else:
                    if key not in compressed:
                        compressed[key] = value
                    elif compressed[key].__repr__ != value.__repr__:
                        compressed_calc[key] = value
                    else:
                        compressed[key] = value

        self.metadata = compressed

    def __str__(self):
        return 'upload %s upload_id%s' % (super().__str__(), self.upload_id)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import numpy as np
import logging
import re
from datetime import datetime
from ase.io import read as ioread
from ase import Atoms

from nomad.units import ureg
from nomad.parsing import FairdiParser
from nomad.parsing.file_parser import TextParser, Quantity
from nomad.datamodel.metainfo.common_dft import Run, Method, System, XCFunctionals,\
    SingleConfigurationCalculation, ScfIteration, Eigenvalues, SamplingMethod, Dos,\
    SpeciesProjectedDos

from wien2kparser.metainfo import m_env
from wien2kparser.metainfo.wien2k import x_wien2k_section_equiv_atoms


class In0Parser(TextParser):
    def __init__(self):
        super().__init__()

    def init_quantities(self):
        self._quantities = [
            Quantity(
                'xc_functional',
                r'(?:TOT|KXC|POT|MULT|COUL|EXCH)\s*([\w ]+)', dtype=str),
            Quantity('fft', r'FFT[\s\S]+?(\d+\s+\d+\s+\d+\s+[\d\.]+)')]


class In1Parser(TextParser):
    def __init__(self):
        super().__init__()

    def init_quantities(self):
        self._quantities = [
            Quantity('wf_switch', r'(WFFIL|WFPRI|ENFIL|SUPWF)'),
            Quantity('rkmax', r'([\d\.]+\s*\d+\s*\d+).+K\-MAX', dtype=np.float64)]


class StructParser(TextParser):
    def __init__(self):
        super().__init__()
        self._units_map = {
            'b': ureg.bohr, 'a': ureg.angstrom}

    def init_quantities(self):
        re_float = r'[\d\.\-]+'
        re_lat = r'\d+\.\d{6}'
        self._quantities = [
            Quantity(
                'lattice',
                r'([\s\S]+?)\n *AT',
                sub_parser=TextParser(quantities=[
                    Quantity('nonequiv_atoms', r'NONEQUIV\.ATOMS\:\s*(\d+)', dtype=int),
                    Quantity('lattice', r'(\w+)\s*LATTICE'),
                    Quantity('calc_mode', r'(N*REL\S*)'),
                    Quantity(
                        'lattice_constants',
                        # fixed precision, sometimes no spaces
                        rf'({re_lat})\s*({re_lat})\s*({re_lat})\s*({re_lat})\s*({re_lat})\s*({re_lat})',
                        dtype=np.dtype(np.float64)),
                    Quantity(
                        'unit',
                        r'unit=(\w)',
                        str_operation=lambda x: self._units_map.get(x, ureg.bohr))])),
            Quantity(
                'atom',
                r'OM\s+\-*\d+\:\s*(X\=[\s\S]+?)LOCAL',
                repeats=True, sub_parser=TextParser(quantities=[
                    Quantity(
                        'positions',
                        rf'X\=({re_float})\s*Y=({re_float})\s*Z=({re_float})',
                        repeats=True, dtype=np.dtype(np.float64)),
                    Quantity('atom_name', r'(\n *[A-Z][a-z]*\d* +)'),
                    Quantity('NPT', r'NPT\s*\=\s*(\d+)', dtype=int),
                    Quantity('R0', r'R0\s*\=\s*(\d+)', dtype=int),
                    Quantity('Z', r'Z\:\s*([\d\.]+)', dtype=np.float64)]))]

    def get_atoms(self):
        '''
        Returns an ASE atoms representation of the structure.
        '''
        if self.mainfile is None:
            return

        try:
            return ioread(self.mainfile, format='struct')
        except Exception:
            # read it from parsed info
            if self.get('lattice') is None:
                return

            lattice_constants = self.get('lattice').get('lattice_constants')
            unit = self.get('lattice').get('unit', ureg.bohr)
            lattice_constants[:3] = (lattice_constants[:3] * unit).to('angstrom').magnitude
            scaled_positions = []
            numbers = []
            for atom in self.get('atom', []):
                positions = atom.get('positions', [])
                scaled_positions.extend(positions)
                numbers.extend([int(atom.get('Z', 0))] * len(positions))

            return Atoms(
                cell=lattice_constants, scaled_positions=scaled_positions, numbers=numbers, pbc=True)


class In2Parser(TextParser):
    def __init__(self):
        super().__init__()

    def init_quantities(self):
        self._quantities = [
            Quantity('switch', r'(TOT|FOR|QTL|EFG|FERMI)'),
            Quantity('emin', r'([\d\.\- ]+)\s*EMIN'),
            Quantity('smearing', r'(GAUSS|ROOT|TEMP|TETRA|ALL)\s*([\d\.]+)'),
            Quantity('gmax', r'([\d\.\-]+)\s*GMAX')]


class DosParser(TextParser):
    def __init__(self):
        super().__init__()

    def init_quantities(self):
        re_f = r'\-*\d+\.\d+'
        self._quantities = [
            Quantity('labels', r'# ENERGY +(.+)', flatten=False),
            Quantity('data', rf'({re_f} +{re_f}.*)', repeats=True, dtype=np.dtype(np.float64))]


class OutParser(TextParser):
    def __init__(self):
        super().__init__()

    def init_quantities(self):
        re_float = r'[\d\.Ee\-\+]+'

        iteration_quantities = [
            Quantity(
                'NATO',
                r'(NATO\s*\:[\s\S]+?)\n *\:',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'nr_of_independent_atoms',
                        r'(\d+)\s*INDEPENDENT', dtype=int),
                    Quantity(
                        'total_atoms',
                        r'(\d+)\s*TOTAL ATOMS IN UNITCELL', dtype=int),
                    Quantity('system_name', r'SUBSTANCE:\s*(.+)', flatten=False)])),
            Quantity(
                'POT',
                r'(POT\s*\:[\s\S]+?)\n *\:',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'potential_option',
                        r'POTENTIAL OPTION\s*(.+)', dtype=str, flatten=False)])),
            Quantity(
                'LAT',
                r'(LAT\s*\:[\s\S]+?)\n *\:',
                sub_parser=TextParser(quantities=[
                    Quantity('lattice_const', r'LATTICE CONSTANTS=\s*([\d\. ]+)')])),
            Quantity(
                'VOL',
                r'(VOL\s*\:[\s\S]+?)\n *\:',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'unit_cell_volume_bohr3',
                        rf'UNIT CELL VOLUME\s*\=\s*({re_float})',
                        dtype=np.float64),
                    Quantity('spinpolarization', r'((?:NON-)*SPINPOLARIZED) CALCULATION')])),
            Quantity(
                'RKM',
                r'(RKM\s*\:[\s\S]+?)\n *\:',
                sub_parser=TextParser(quantities=[
                    Quantity('matrix_size', r'MATRIX SIZE\s*(\d+)', dtype=int),
                    Quantity('LOs', r'LOs:\s*(\d+)', dtype=int),
                    Quantity('rkm', r'RKM\=\s*([\d\.]+)', dtype=np.float64)])),
            Quantity(
                'KPT',
                r'(KPT\s*\:[\s\S]+?)\n *\:',
                sub_parser=TextParser(quantities=[Quantity(
                    'nr_kpts', r'NUMBER OF K-POINTS:\s*(\d+)', dtype=int)])),
            Quantity(
                'GAP',
                r'(GAP\s*\:[\s\S]+?)\n *\:',
                sub_parser=TextParser(quantities=[Quantity(
                    'ene_gap', rf'({re_float})\s*Ry', dtype=np.float64, unit=ureg.rydberg)])),
            Quantity(
                'NOE',
                r'(NOE\s*\:[\s\S]+?)\n *\:',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'noe',
                        rf'NUMBER OF ELECTRONS\s*\=\s*({re_float})',
                        dtype=np.float64)])),
            Quantity(
                'FER',
                r'(FER\s*\:[\s\S]+?)\n *\:',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'energy_reference_fermi_iteration',
                        rf'F E R M I \- ENERGY.+?\=\s*([\d\.\-\+Ee ]+)',
                        str_operation=lambda x: [float(v) for v in x.strip().split()] * ureg.rydberg,
                        convert=False)])),
            Quantity(
                'GMA',
                r'(GMA\s*\:[\s\S]+?)\n *\:',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'cutoff',
                        rf'POTENTIAL AND CHARGE CUT\-OFF\s*({re_float})',
                        dtype=np.float64)])),
            Quantity(
                'POSi',
                r'(POS\d+\:[\s\S]+?)\n *\:',
                repeats=True, sub_parser=TextParser(quantities=[
                    Quantity('atom_mult', r'MULT.*?\s*\=\s*(\d+)', dtype=int),
                    Quantity(
                        'position',
                        rf'POSITION\s*\=\s*({re_float}\s*{re_float}\s*{re_float})',
                        dtype=np.float64)])),
            Quantity(
                'CHAi',
                r'(CHA\d+\:[\s\S]+?)\n *\:',
                repeats=True, sub_parser=TextParser(quantities=[
                    Quantity(
                        'tot_val_charge_cell',
                        rf'TOTAL .+?CHARGE INSIDE.+?\=\s*({re_float})',
                        dtype=np.float64)])),
            Quantity(
                'SUM',
                r'(SUM\s*\:[\s\S]+?)\n *\:',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'energy_sum_eigenvalues_scf_iteration',
                        rf'SUM OF EIGENVALUES\s*\=\s*({re_float})',
                        dtype=np.float64, unit=ureg.rydberg)])),
            Quantity(
                'RTOi',
                rf'RTO\d+\:\s*\d+\s*({re_float})\s*({re_float})\s*({re_float})\s*({re_float})\s*',
                dtype=np.dtype(np.float64), repeats=True),
            Quantity(
                'NTO',
                r'(NTO\s*\:[\s\S]+?)\n *\:',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'tot_int_charge_nm',
                        rf'CHARGE\s*\=\s*({re_float})', dtype=np.float64)])),
            Quantity(
                'NTOi',
                r'(NTO\d+\:[\s\S]+?)\n *\:',
                repeats=True, sub_parser=TextParser(quantities=[
                    Quantity(
                        'tot_charge_in_sphere_nm',
                        rf'CHARGE.+\=\s*({re_float})', dtype=np.float64)])),
            Quantity(
                'DTOi',
                r'(DTO\d+\:[\s\S]+?)\n *\:',
                repeats=True, sub_parser=TextParser(quantities=[
                    Quantity(
                        'tot_diff_charge',
                        rf'TOTAL\s*DIFFERENCE CHARGE.+\=\s*({re_float})', dtype=np.float64)])),
            Quantity(
                'DIS',
                r'(DIS\s*\:[\s\S]+?)\n *\:',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'charge_distance',
                        rf'CHARGE DISTANCE.+\)\s*({re_float})', dtype=np.float64)])),
            Quantity(
                'CTO',
                r'(CTO\s*\:[\s\S]+?)\n *\:',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'tot_int_charge',
                        rf'CHARGE\s*\=\s*({re_float})', dtype=np.float64)])),
            Quantity(
                'CTOi',
                r'(CTO\d+\:[\s\S]+?)\n *\:',
                repeats=True, sub_parser=TextParser(quantities=[
                    Quantity(
                        'tot_charge_in_sphere',
                        rf'TOTAL\s*CHARGE IN SPHERE.+\=\s*({re_float})', dtype=np.float64)])),
            Quantity(
                'NECi',
                rf'NEC\d+\:\s*NUCLEAR AND ELECTRONIC CHARGE\s*({re_float})\s*({re_float})',
                dtype=np.dtype(np.float64), repeats=True),
            Quantity(
                'MMINT',
                r'(MMINT\s*\:[\s\S]+?)\n *\:',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'mmint',
                        rf'MAGNETIC MOMENT IN INTERSTITIAL\s*\=\s*({re_float})', dtype=np.float64)])),
            Quantity(
                'MMIi',
                r'(MMI\d+\:[\s\S]+?)\n *\:',
                repeats=True, sub_parser=TextParser(quantities=[
                    Quantity(
                        'mmi',
                        rf'MAGNETIC MOMENT IN SPHERE\s*\=\s*({re_float})', dtype=np.float64)])),
            Quantity(
                'MMTOT',
                r'(MMTOT\s*\:[\s\S]+?)\n *\:',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'mmtot',
                        rf' MAGNETIC MOMENT IN CELL\s*\=\s*({re_float})', dtype=np.float64)])),
            Quantity(
                'ENE',
                r'(ENE\s*\:[\s\S]+?)\n',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'energy_total_scf_iteration',
                        rf'TOTAL ENERGY IN Ry\s*\=\s*({re_float})',
                        dtype=np.float64, unit=ureg.rydberg)])),
            Quantity(
                'FORi',
                rf'FOR\d+\:\s*\d+\.ATOM\s*({re_float}\s*{re_float}\s*{re_float}\s*{re_float})',
                repeats=True, dtype=np.dtype(np.float64)),
            Quantity(
                'FGLi',
                rf'FGL\d+\:\s*\d+\.ATOM\s*({re_float}\s*{re_float}\s*{re_float})',
                repeats=True, dtype=np.dtype(np.float64))]

        self._quantities = [
            Quantity(
                'version',
                r'LABEL\d+\:\s*using WIEN2k_(\S+) \(Release ([\d\/]+)\)', flatten=False),
            Quantity(
                'start_date',
                r'LABEL\d+\:\s*on .+ at \w+ (\w+ \d+ \d\d\:\d\d\:\d\d)\s*\w*\s*(\d+)',
                flatten=False),
            Quantity(
                'iteration',
                r'\d+\:\s*(\d+\.\s* ITERATION[\s\S]+?)(?:\:ITE|\Z)',
                repeats=True, sub_parser=TextParser(quantities=iteration_quantities))]


class Wien2kParser(FairdiParser):
    def __init__(self):
        super().__init__(
            name='parsers/wien2k', code_name='WIEN2k', code_homepage='http://www.wien2k.at/',
            mainfile_name_re=r'.*\.scf$',
            mainfile_alternative=True,
            mainfile_contents_re=r'\s*---------\s*:ITE[0-9]+:\s*[0-9]+\.\s*ITERATION\s*---------')

        # TODO complete list and verify if consistent with current implementation
        # http://www.wien2k.at/reg_user/textbooks/usersguide.pdf
        # implement libxc compatibility
        self._xc_functional_map = {
            '5': ['LDA_X', 'LDA_C_PW'],
            '6': ['HF_X'],
            '11': ['GGA_X_WC', 'GGA_C_PBE'],
            '13': ['GGA_X_PBE', 'GGA_C_PBE'],
            '17': ['GGA_X_PW91'],
            '18': ['HYB_GGA_XC_B3PW91'],
            '19': ['GGA_X_PBE_SOL', 'GGA_C_PBE_SOL'],
            '24': ['GGA_X_B88', 'GGA_C_LYP'],
            '28': ['MGGA_X_TB09', 'LDA_C_PW'],
            '27': ['MGGA_X_TPSS', 'MGGA_C_TPSS'],
            '29': ['MGGA_C_REVTPSS, GGA_C_REGTPSS'],
            '46': ['GGA_X_HTBS'],
            '47': ['HYB_GGA_XC_B3LYP'],
            'XC_LDA': ['LDA_X', 'LDA_C_PW'],
            'XC_PBE': ['GGA_X_PBE', 'GGA_C_PBE'],
            'XC_WC': ['GGA_X_WC', 'GGA_C_PBE'],
            'XC_PBESOL': ['GGA_X_PBE_SOL', 'GGA_C_PBE_SOL'],
            'XC_B3PW91': ['HYB_GGA_XC_B3PW91'],
            'XC_B3LYP': ['HYB_GGA_XC_B3LYP'],
            # see issue #1, not sure where TB09 comes from
            'XC_MBJ': ['MGGA_X_TB09', 'LDA_C_PW'],
            'XC_LMBJ': ['LDA_X', 'LDA_C_PW'],
            'XC_TPSS': ['MGGA_X_TPSS', 'MGGA_C_TPSS'],
            'XC_REVTPSS': ['MGGA_C_REVTPSS, GGA_C_REGTPSS'],
            'XC_MGGA_MS': ['MGGA_X_MS', 'MGGA_C_MS'],
            'XC_MVS': ['MGGA_X_MVS', 'MGGA_C_MVS'],
            'XC_MBEEF': ['MGGA_X_MBEEF', 'GGA_C_PBE_SOL'],
            'XC_SCAN': ['MGGA_X_SCAN', 'MGGA_C_SCAN'],
            'XC_SCANL': ['MGGA_X_SCANL', 'MGGA_C_SCANL'],
            'XC_RSCAN': ['MGGA_X_RSCAN', 'MGGA_C_RSCAN'],
            'XC_R2SCAN': ['MGGA_X_R2SCAN', 'MGGA_C_R2SCAN'],
            'XC_TM': ['MGGA_X_TM', 'MGGA_C_TM'],
            'EC_PW91': ['GGA_X_PW91'],
            'VC_PW91': ['GGA_X_PW91'],
            'EX_B88': ['GGA_X_B88'],
            'VX_B88': ['GGA_X_B88'],
            'EC_LYP': ['GGA_C_LYP'],
            'VC_LYP': ['GGA_C_LYP'],
            'XC_HTBS': ['GGA_X_HTBS'],
            'EX_LDA': ['HF_X'],
            'VX_LDA': ['HF_X']}

        self._metainfo_env = m_env
        self.out_parser = OutParser()
        self.in0_parser = In0Parser()
        self.in1_parser = In1Parser()
        self.in2_parser = In2Parser()
        self.struct_parser = StructParser()
        self.dos_parser = DosParser()

    def init_parser(self):
        self.out_parser.mainfile = self.filepath
        self.out_parser.logger = self.logger

    def get_wien2k_file(self, ext, multiple=False):
        paths = [p for p in os.listdir(self.maindir) if re.match(r'.*%s$' % ext, p)]
        if not paths:
            return [] if multiple else None
        elif len(paths) == 1:
            path = os.path.join(self.maindir, paths[0])
            return [path] if multiple else path
        else:
            prefix = os.path.basename(self.filepath).rsplit('.', 1)[0]
            paths = [os.path.join(self.maindir, p) for p in paths if p.startswith(prefix)]
            if not paths:
                return [] if multiple else None
            return paths if multiple else paths[0]

    def get_nspin(self):
        return 2 if self.out_parser.get('iteration', [{}])[0].get('VOL', {}).get('spinpolarization') == 'SPINPOLARIZED' else 1

    def get_kpoints(self):
        k_list_file = self.get_wien2k_file('klist')
        if k_list_file is None:
            return None

        kpoints = []
        weights = []
        with open(k_list_file) as f:
            while True:
                line = f.readline()
                if not line or 'END' in line:
                    break
                try:
                    line = np.array(line.split()[:6], dtype=float)
                except Exception:
                    continue
                kpoints.append(line[1:4] / line[4])
                weights.append(line[5])
        weights /= sum(weights)

        return kpoints, weights

    def get_eigenvalues(self):
        nspin = self.get_nspin()
        if nspin == 1:
            files = self.get_wien2k_file(r'energy\_\d+', multiple=True)
            if not files:
                files = [self.get_wien2k_file('energy')]
        else:
            files = []
            for spin in ['up', 'dn']:
                files_spin = self.get_wien2k_file(r'energy%s\_\d+' % spin, multiple=True)
                if not files_spin:
                    files_spin = [self.get_wien2k_file('energy%s' % spin)]
                files.extend(files_spin)
        if None in files:
            return

        re_k = r'\-*\d\.\d+E[\-\+]\d\d'
        re_kpoint = re.compile(rf'\s*({re_k})\s*({re_k})\s*({re_k})\w*\s*(\d+)\s*\d+\s*\d+\s*([\d\.]+)\s*')
        re_eigenvalue = re.compile(r'\s*\d+ +(\-*\d+\.\d+[E\-\+\d]+) *\n*')
        kpoints, eigenvalues, multiplicity, index = [], [], [], []
        for file_i in files:
            with open(file_i) as f:
                while True:
                    line = f.readline()
                    if not line:
                        break
                    if 'energydn' not in file_i:
                        kpoint = re_kpoint.match(line)
                        if kpoint:
                            kpoints.append(kpoint.groups()[:3])
                            index.append(kpoint.group(4))
                            multiplicity.append(kpoint.group(5))
                            continue
                    eigenvalue = re_eigenvalue.match(line)
                    if eigenvalue:
                        eigenvalues.append(eigenvalue.group(1))
        try:
            eigenvalues = np.array(eigenvalues, dtype=np.dtype(np.float64))
            kpoints = np.array(kpoints, dtype=np.dtype(np.float64))
            multiplicity = np.array(multiplicity, dtype=np.dtype(np.float64))
            eigenvalues = np.reshape(eigenvalues, (
                nspin, len(kpoints), len(eigenvalues) // (nspin * len(kpoints))))
            # sort data wrt kpoint index
            index = np.argsort(index)
            kpoints = kpoints[index]
            multiplicity = multiplicity[index]
            for spin in range(nspin):
                eigenvalues[spin] = eigenvalues[spin][index]
            return eigenvalues, kpoints, multiplicity
        except Exception:
            self.logger.error('Error reading eigenvalues.')
            return

    def get_dos(self):
        # dos (projections and total) are printed in dos1, dos2....
        nspin = self.get_nspin()
        if nspin == 1:
            files = self.get_wien2k_file(r'dos\d+', multiple=True)
            files.sort()
        else:
            files = []
            for spin in ['up', 'dn']:
                files_spin = self.get_wien2k_file(r'dos\d+%s' % spin, multiple=True)
                files_spin.sort()
                files.extend(files_spin)

        if not files:
            return

        dos = []
        labels = []
        for file_i in files:
            self.dos_parser.mainfile = file_i
            data = self.dos_parser.get('data')
            if data is None:
                continue
            if not file_i.endswith('dn'):
                labels.extend(self.dos_parser.get('labels', '').split())
            data = np.transpose(data)
            energy = data[0]
            dos.append(data[1:])

        try:
            dos = np.vstack(dos)
            dos = np.reshape(dos, (nspin, len(dos) // nspin, len(dos[0])))
            dos = np.transpose(dos, axes=(1, 0, 2))
            total_dos = []
            partial_dos = []
            for n in range(len(dos)):
                # wien2k may not uniformly print out projections, we get only total projections
                # on the species (independent atoms) denoted by the header N:total
                if labels[n].endswith(':total'):
                    partial_dos.append(dos[n])
                elif labels[n] == 'total-DOS' or labels[n] == 'TOTAL':
                    # TODO determine if total dos is always the last column
                    total_dos = dos[n]
            if len(partial_dos) > 0:
                partial_dos = np.transpose(partial_dos, axes=(1, 0, 2))
            return energy, total_dos, partial_dos

        except Exception:
            self.logger.error('Error reading dos.')
            return

    def parse_scc(self):
        if self.out_parser.get('iteration') is None:
            return

        sec_scc = self.archive.section_run[0].m_create(SingleConfigurationCalculation)

        for iteration in self.out_parser.get('iteration'):
            sec_scf = sec_scc.m_create(ScfIteration)
            # sec_scf.energy_total_scf_iteration =
            for key in iteration.keys():
                if iteration.get(key) is None:
                    continue
                elif key == 'FORi':
                    forces = np.transpose(iteration.get(key))
                    sec_scf.x_wien2k_for = np.transpose(forces[1:4])
                    sec_scf.x_wien2k_for_abs = forces[0]
                elif key == 'FGLi':
                    sec_scf.x_wien2k_for_gl = iteration.get(key)
                elif key == 'MMIi':
                    sec_scf.x_wien2k_mmi = [mm.mmi for mm in iteration.get(key)]
                elif key == 'NECi':
                    charge = np.transpose(iteration.get(key))
                    sec_scf.x_wien2k_nuclear_charge = charge[0]
                    sec_scf.x_wien2k_electronic_charge = charge[1]
                elif key == 'CTOi':
                    charge = [c.tot_charge_in_sphere for c in iteration.get(key)]
                    sec_scf.x_wien2k_tot_charge_in_sphere = charge
                elif key == 'DTOi':
                    charge = [c.tot_diff_charge for c in iteration.get(key)]
                    sec_scf.x_wien2k_tot_diff_charge = charge
                elif key == 'NTOi':
                    charge = [c.tot_charge_in_sphere_nm for c in iteration.get(key)]
                    sec_scf.x_wien2k_tot_charge_in_sphere_nm = charge
                elif key == 'RTOi':
                    density = np.transpose(iteration.get(key))
                    sec_scf.x_wien2k_density_at_nucleus_valence = density[0]
                    sec_scf.x_wien2k_density_at_nucleus_semicore = density[1]
                    sec_scf.x_wien2k_density_at_nucleus_core = density[2]
                    sec_scf.x_wien2k_density_at_nucleus_tot = density[3]
                elif key == 'CHAi':
                    charge = [c.tot_val_charge_cell for c in iteration.get(key)]
                    sec_scf.x_wien2k_tot_val_charge_sphere = charge
                elif key == 'POSi':
                    mult = [p.atom_mult for p in iteration.get(key)]
                    sec_scf.x_wien2k_atom_mult = mult
                else:
                    for sub_key, val in iteration.get(key, {}).items():
                        sub_key = sub_key if sub_key.startswith('energy_') else 'x_wien2k_%s' % sub_key
                        setattr(sec_scf, sub_key, val)

        # write final iteration values to scc
        sec_scc.energy_total = sec_scf.energy_total_scf_iteration

        if sec_scf.x_wien2k_for_gl is not None:
            forces = []
            for n, force in enumerate(sec_scf.x_wien2k_for_gl):
                forces.extend([force] * sec_scf.x_wien2k_atom_mult[n])
            sec_scc.atom_forces = forces * (ureg.mRy / ureg.bohr)

        sec_scc.energy_reference_fermi = sec_scf.energy_reference_fermi_iteration

        # eigenvalues
        eigenvalues = self.get_eigenvalues()
        if eigenvalues is not None:
            sec_eigenvalues = sec_scc.m_create(Eigenvalues)
            sec_eigenvalues.eigenvalues_values = eigenvalues[0] * ureg.rydberg
            sec_eigenvalues.eigenvalues_kpoints = eigenvalues[1]
            sec_eigenvalues.eigenvalues_kpoints_multiplicity = eigenvalues[2]

        # dos
        dos = self.get_dos()
        if dos is not None:
            # total dos
            if len(dos[1]) > 0:
                sec_dos = sec_scc.m_create(Dos)
                sec_dos.dos_values = (dos[1] * (1 / ureg.rydberg)).to('1 / J').magnitude
                sec_dos.dos_energies = dos[0] * ureg.rydberg

            # projected dos
            if len(dos[2]) > 0:
                sec_dos = sec_scc.m_create(SpeciesProjectedDos)
                sec_dos.species_projected_dos_energies = dos[0] * ureg.rydberg
                sec_dos.species_projected_dos_values_total = (dos[2] * (1 / ureg.rydberg)).to('1 / J').magnitude
                labels = [a.atom_name for a in self.struct_parser.get('atom', [])]
                sec_dos.species_projected_dos_species_label = labels

    def parse_system(self):
        sec_system = self.archive.section_run[0].m_create(System)

        self.struct_parser.mainfile = self.get_wien2k_file('struct')
        for key, val in self.struct_parser.get('lattice', {}).items():
            setattr(sec_system, 'x_wien2k_%s' % key, val)

        for atom in self.struct_parser.get('atom', []):
            sec_atom = sec_system.m_create(x_wien2k_section_equiv_atoms)
            for key, val in atom.items():
                setattr(sec_atom, 'x_wien2k_%s' % key, val)

        atoms = self.struct_parser.get_atoms()
        if atoms is None:
            return

        sec_system.lattice_vectors = np.array(atoms.get_cell()) * ureg.angstrom
        sec_system.atom_positions = atoms.get_positions() * ureg.angstrom
        sec_system.atom_labels = atoms.get_chemical_symbols()
        sec_system.configuration_periodic_dimensions = atoms.get_pbc()

    def parse_method(self):
        sec_method = self.archive.section_run[0].m_create(Method)

        sec_method.number_of_spin_channels = self.get_nspin()

        # read functional settings from in0 file
        self.in0_parser.mainfile = self.get_wien2k_file('in0')

        # better to read it from scf?
        xc_functional = self.in0_parser.get('xc_functional', None)
        xc_functional = xc_functional if isinstance(xc_functional, list) else [xc_functional]
        for name in xc_functional:
            functionals = self._xc_functional_map.get(name)
            if functionals is None:
                self.logger.warn('Cannot resolve XC functional.')
                continue
            for functional in functionals:
                sec_xc_functional = sec_method.m_create(XCFunctionals)
                sec_xc_functional.XC_functional_name = functional

        fft = self.in0_parser.get('fft')
        if fft is not None:
            sec_method.x_wien2k_ifft = fft[:3]
            sec_method.x_wien2k_ifft_factor = fft[3]

        # read cut off settings from in1
        in1_file = self.get_wien2k_file('in1')
        if in1_file is None:
            in1_file = self.get_wien2k_file('in1c')
        self.in1_parser.mainfile = in1_file
        for key, val in self.in1_parser.items():
            if val is not None:
                setattr(sec_method, 'x_wien2k_%s' % key, val)

        # read integration data from in2 file
        in2_file = self.get_wien2k_file('1n2')
        if in2_file is None:
            in2_file = self.get_wien2k_file('in2c')
        self.in2_parser.mainfile = in2_file

        for key in ['gmax', 'switch']:
            val = self.in2_parser.get(key)
            if val is not None:
                setattr(sec_method, 'x_wien2k_in2_%s' % key, val)

        emin_keys = ['emin', 'ne', 'espermin', 'esper0']
        for n, val in enumerate(self.in2_parser.get('emin', [])):
            if n < 4:
                setattr(sec_method, 'x_wien2k_in2_%s' % emin_keys[n], val)

        smearing, width = self.in2_parser.get('smearing', [None, None])
        if smearing is not None:
            if smearing.startswith('GAUSS'):
                smearing = 'gaussian'
            elif smearing.startswith('TEMP'):
                smearing = 'fermi'
            elif smearing.startswith('TETRA'):
                smearing = 'tetrahedra'
            sec_method.smearing_kind = smearing
            sec_method.smearing_width = (float(width) * ureg.rydberg).to('joule').magnitude

        # read kpoints from klist
        kpoints = self.get_kpoints()
        if kpoints is not None:
            sec_method.k_mesh_points = kpoints[0]
            sec_method.k_mesh_weights = kpoints[1]

    def parse(self, filepath, archive, logger):
        self.filepath = os.path.abspath(filepath)
        self.archive = archive
        self.maindir = os.path.dirname(self.filepath)
        self.logger = logger if logger is not None else logging.getLogger(__name__)

        self.init_parser()

        sec_run = self.archive.m_create(Run)

        sec_run.program_name = 'WIEN2k'
        sec_run.program_basis_set_type = '(L)APW+lo'
        sec_run.program_version = self.out_parser.get('version', '')
        start_date = self.out_parser.get('start_date')
        if start_date is not None:
            # TODO resolve proper timezone
            dt = datetime.strptime(start_date, '%b %d %H:%M:%S %Y') - datetime.utcfromtimestamp(0)
            sec_run.time_run_date_start = dt.total_seconds()

        # TODO implement geometry optimization
        sec_sampling = sec_run.m_create(SamplingMethod)
        sec_sampling.sampling_method = 'single_point'

        self.parse_method()

        self.parse_system()

        self.parse_scc()



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

''' An example metainfo package. '''

import numpy as np
from datetime import datetime

from nomad.units import ureg
from nomad.metainfo import (
    MSection, MCategory, Section, Quantity, Package, SubSection, MEnum,
    Datetime, constraint)

m_package = Package(links=['https://nomad-lab.eu/prod/rae/docs/metainfo.html'])


class SystemHash(MCategory):
    ''' All quantities that contribute to what makes a system unique. '''


class Parsing(MSection):
    ''' All data that describes the NOMAD parsing of this run.

    Quantities can also be documented like this:

    Args:
        parser_name: 'Name of the used parser'
        parser_version: 'Version of the used parser'
    '''

    parser_name = Quantity(type=str)
    parser_version = Quantity(type=str)
    nomad_version = Quantity(type=str, default='latest')
    warnings = Quantity(type=str, shape=['0..*'])
    parse_time = Quantity(type=Datetime)


class System(MSection):
    ''' All data that describes a simulated system. '''

    n_atoms = Quantity(
        type=int, derived=lambda system: len(system.atom_labels),
        description='Number of atoms in the simulated system.')

    atom_labels = Quantity(
        type=str, shape=['n_atoms'], categories=[SystemHash],
        description='The atoms in the simulated systems.')

    atom_positions = Quantity(
        type=np.dtype('f'), shape=['n_atoms', 3], unit=ureg.m, categories=[SystemHash],
        description='The atom positions in the simulated system.')

    lattice_vectors = Quantity(
        type=np.dtype('f'), shape=[3, 3], unit=ureg.m, categories=[SystemHash],
        aliases=['unit_cell'],
        description='The lattice vectors of the simulated unit cell.')

    periodic_dimensions = Quantity(
        type=bool, shape=[3], default=[False, False, False], categories=[SystemHash],
        description='A vector of booleans indicating in which dimensions the unit cell is repeated.')

    system_type = Quantity(type=str)


class SCC(MSection):

    energy_total = Quantity(type=float, default=0.0, unit=ureg.J)
    energy_total_0 = Quantity(type=np.dtype(np.float32), default=0.0, unit=ureg.J)
    an_int = Quantity(type=np.dtype(np.int32))

    system = Quantity(type=System, description='The system that this calculation is based on.')


class Run(MSection):
    ''' All data that belongs to a single code run. '''

    code_name = Quantity(type=str, description='The name of the code that was run.')
    code_version = Quantity(type=str, description='The version of the code that was run.')

    parsing = SubSection(sub_section=Parsing)
    systems = SubSection(sub_section=System, repeats=True)
    sccs = SubSection(sub_section=SCC, repeats=True)

    @constraint
    def one_scc_per_system(self):
        assert self.m_sub_section_count(Run.systems) == self.m_sub_section_count(Run.sccs),\
            'Numbers of system does not match numbers of calculations.'


class VaspRun(Run):
    ''' All VASP specific quantities for section Run. '''
    m_def = Section(extends_base_section=True)

    x_vasp_raw_format = Quantity(
        type=MEnum(['xml', 'outcar']),
        description='The file format of the parsed VASP mainfile.')


if __name__ == '__main__':
    # Demonstration of how to reflect on the definitions

    # All definitions are metainfo data themselves, and they can be accessed like any other
    # metainfo data. E.g. all section definitions are sections themselves.

    # To get quantities of a given section
    print(Run.m_def.m_get_sub_sections(Section.quantities))

    # Or all Sections in the package
    print(m_package.m_get_sub_sections(Package.section_definitions))

    # There are also some definition specific helper methods.
    # For example to get all attributes (Quantities and possible sub-sections) of a section.
    print(Run.m_def.all_properties)

    # Demonstration on how to use the definitions, e.g. to create a run with system:
    run = Run()
    run.code_name = 'VASP'
    run.code_version = '1.0.0'

    parsing = run.m_create(Parsing)
    parsing.parse_time = datetime.now()

    run.m_as(VaspRun).x_vasp_raw_format = 'outcar'
    # The same as
    run.x_vasp_raw_format = 'outcar'  # type: ignore

    system = run.m_create(System)
    system.atom_labels = ['H', 'H', 'O']

    calc = run.m_create(SCC)
    calc.energy_total = 1.23e-10
    calc.system = system

    # Or to read data from existing metainfo data:
    print(system.atom_labels)
    print(system.n_atoms)

    # To validate dimensions and custom constraints
    print('errors: %s' % run.m_all_validate())

    # To serialize the data:
    serializable = run.m_to_dict()
    # or
    print(run.m_to_json(indent=2))

    # To deserialize data
    run = Run.m_from_dict(serializable)
    print(run.sccs[0].system)

    # print(m_package.m_to_json(indent=2))  # type: ignore, pylint: disable=undefined-variable
